#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
This module contains classes for XML Schema elements, complex types and model groups.
"""
import warnings
from copy import copy as _copy
from decimal import Decimal
from types import GeneratorType
from collections.abc import Iterator, MutableSequence
from typing import TYPE_CHECKING, cast, Any, Optional, Union
from xml.etree.ElementTree import Element, ParseError

from elementpath import XPath2Parser, ElementPathError, XPathContext, XPathToken, \
    LazyElementNode, SchemaElementNode, build_schema_node_tree
from elementpath.datatypes import AbstractDateTime, Duration
from elementpath.xpath_nodes import EtreeElementNode

import xmlschema.names as nm
from xmlschema.exceptions import XMLSchemaTypeError, XMLSchemaValueError, \
    XMLResourceParseError
from xmlschema.aliases import ElementType, BaseXsdType, SchemaElementType, \
    ModelParticleType, ComponentClassType, DecodeType, DecodedValueType
from xmlschema.translation import gettext as _
from xmlschema.utils.etree import iter_schema_location_hints, iter_schema_namespaces
from xmlschema.utils.decoding import Empty, raw_encode_attributes, strictly_equal
from xmlschema.utils.qnames import get_qname
from xmlschema.arguments import XSD_VALIDATION_MODES
from xmlschema import dataobjects
from xmlschema.converters import ElementData
from xmlschema.xpath import XMLSchemaProxy, ElementPathMixin, XPathElement
from xmlschema.caching import schema_cache

from .exceptions import XMLSchemaValidationError, XMLSchemaParseError, \
    XMLSchemaStopValidation, XMLSchemaTypeTableWarning
from .validation import ValidationContext, DecodeContext, EncodeContext, ValidationMixin
from .helpers import parse_xsd_derivation, parse_xpath_default_namespace
from .xsdbase import XSD_TYPE_DERIVATIONS, XSD_ELEMENT_DERIVATIONS, XsdComponent
from .particles import ParticleMixin, OccursCalculator
from .identities import XsdIdentity, XsdKeyref, KeyrefCounter, FieldValueSelector
from .simple_types import XsdSimpleType
from .attributes import XsdAttribute
from .wildcards import XsdAnyElement

if TYPE_CHECKING:
    from .attributes import XsdAttributeGroup  # noqa: F401
    from .groups import XsdGroup  # noqa: F401

DataBindingType = Union[type['dataobjects.DataElement'], 'dataobjects.DataBindingMeta']


class XsdElement(XsdComponent, ParticleMixin,
                 ElementPathMixin[SchemaElementType],
                 ValidationMixin[ElementType, Any]):
    """
    Class for XSD 1.0 *element* declarations.

    ..  <element
          abstract = boolean : false
          block = (#all | List of (extension | restriction | substitution))
          default = string
          final = (#all | List of (extension | restriction))
          fixed = string
          form = (qualified | unqualified)
          id = ID
          maxOccurs = (nonNegativeInteger | unbounded)  : 1
          minOccurs = nonNegativeInteger : 1
          name = NCName
          nillable = boolean : false
          ref = QName
          substitutionGroup = QName
          type = QName
          {any attributes with non-schema namespace . . .}>
          Content: (annotation?, ((simpleType | complexType)?, (unique | key | keyref)*))
        </element>
    """
    name: str
    local_name: str
    qualified_name: str
    prefixed_name: str
    target_namespace: str

    parent: Optional['XsdGroup']
    ref: Optional['XsdElement']

    attributes: 'XsdAttributeGroup'
    """The group of the attributes associated with the element."""

    content: Union[tuple[()], 'XsdGroup']

    abstract: bool = False
    """
    Defines whether the element can be used in an instance document. An abstract
    element must be global and can still be the head of a substitution group.
    """

    nillable: bool = False
    """
    Defines whether the element content is nillable using xsi:nil="true" as attribute.
    """

    form: Optional[str] = None
    qualified: bool = False
    """
    The effective form for the element. If `True` the element name is qualified by a
    braced namespace URI as prefix. The name of a global element is always qualified.
    """

    default: Optional[str] = None
    """The default value of the element if its content is a simple type."""

    fixed: Optional[str] = None
    """The fixed value of the element if its content is a simple type."""

    substitution_group: Optional[str] = None

    substitutes: set[str] | tuple[()] = ()
    identities: list[XsdIdentity]
    selected_by: set[XsdIdentity]
    xsi_types: set[BaseXsdType]
    alternatives: Union[tuple[()], list['XsdAlternative']] = ()
    inheritable: Union[tuple[()], dict[str, XsdAttribute]] = ()

    _ADMITTED_TAGS = nm.XSD_ELEMENT,
    _block: Optional[str] = None
    _final: Optional[str] = None
    _built: bool | None
    binding: Optional[DataBindingType] = None

    __slots__ = ('type', 'selected_by', 'xsi_types', 'identities',
                 'content', 'attributes', 'min_occurs', 'max_occurs')

    def __repr__(self) -> str:
        return '%s(%s=%r, occurs=%r)' % (
            self.__class__.__name__,
            'name' if self.ref is None else 'ref',
            self.prefixed_name,
            list(self.occurs)
        )

    def _set_type(self, value: BaseXsdType) -> None:
        self.type: BaseXsdType = value
        """The XSD simpleType or complexType of the element."""

        if isinstance(value, XsdSimpleType):
            self.attributes = self.builders.create_empty_attribute_group(self)
            self.content = ()
        else:
            self.attributes = value.attributes
            if isinstance(value.content, XsdSimpleType):
                self.content = ()
            else:
                self.content = value.content

    def __iter__(self) -> Iterator[SchemaElementType]:
        if self.content:
            yield from self.content.elements

    def build(self) -> None:
        if self._built is False:
            self._built = None
            try:
                self._parse()
                self._built = True
            finally:
                if self._built is None:
                    self._built = False

    def _parse(self) -> None:
        if self._built is not None and isinstance(self.parent, MutableSequence):
            return

        self.min_occurs = self.max_occurs = 1
        self.selected_by = set()
        self.xsi_types = set()

        self._parse_particle(self.elem)
        self._parse_attributes()

        if self.ref is None:
            self._parse_type()
            self._parse_constraints()

            if self.parent is None:
                self.substitutes = set()
                if 'substitutionGroup' in self.elem.attrib:
                    self._parse_substitution_group(self.elem.attrib['substitutionGroup'])

        self._built = True

    def _parse_attributes(self) -> None:
        attrib = self.elem.attrib
        if self._parse_reference():
            try:
                xsd_element: XsdElement = self.maps.elements[self.name]
            except KeyError:
                self._set_type(self.maps.any_type)
                self.parse_error(_('unknown element %r') % self.name)
            else:
                self.ref = xsd_element
                self._set_type(xsd_element.type)
                self.target_namespace = xsd_element.target_namespace
                self.abstract = xsd_element.abstract
                self.nillable = xsd_element.nillable
                self.qualified = xsd_element.qualified
                self.substitutes = xsd_element.substitutes
                self.form = xsd_element.form
                self.default = xsd_element.default
                self.fixed = xsd_element.fixed
                self.substitution_group = xsd_element.substitution_group
                self.identities = xsd_element.identities
                self.alternatives = xsd_element.alternatives
                self.selected_by = xsd_element.selected_by
                self.xsi_types = xsd_element.xsi_types

            for attr_name in ('type', 'nillable', 'default', 'fixed', 'form',
                              'block', 'abstract', 'final', 'substitutionGroup'):
                if attr_name in attrib:
                    msg = _("attribute {!r} is not allowed when element reference is used")
                    self.parse_error(msg.format(attr_name))
            return

        if 'form' in attrib:
            self.form = attrib['form']
            if self.form == 'qualified':
                self.qualified = True
        elif self.schema.element_form_default == 'qualified':
            self.qualified = True

        try:
            if self.parent is None or self.qualified:
                self.name = get_qname(self.target_namespace, attrib['name'])
            else:
                self.name = attrib['name']
        except KeyError:
            pass

        if 'abstract' in attrib:
            if self.parent is not None:
                msg = _("local scope elements cannot have abstract attribute")
                self.parse_error(msg)
            if attrib['abstract'].strip() in ('true', '1'):
                self.abstract = True

        if 'block' in attrib:
            self._block = parse_xsd_derivation(
                self.elem, 'block', XSD_ELEMENT_DERIVATIONS, self
            )

        if 'nillable' in attrib and attrib['nillable'].strip() in ('true', '1'):
            self.nillable = True

        if self.parent is None:
            if 'final' in attrib:
                self._final = parse_xsd_derivation(
                    self.elem, 'final', XSD_TYPE_DERIVATIONS, self
                )

            for attr_name in ('ref', 'form', 'minOccurs', 'maxOccurs'):
                if attr_name in attrib:
                    msg = _("attribute {!r} is not allowed in a global element declaration")
                    self.parse_error(msg.format(attr_name))
        else:
            for attr_name in ('final', 'substitutionGroup'):
                if attr_name in attrib:
                    msg = _("attribute {!r} not allowed in a local element declaration")
                    self.parse_error(msg.format(attr_name))

    def _parse_type(self) -> None:
        type_name = self.elem.get('type')
        if type_name is not None:
            try:
                extended_name = self.schema.resolve_qname(type_name)
            except (KeyError, ValueError, RuntimeError) as err:
                self.parse_error(err)
                self._set_type(self.maps.any_type)
            else:
                if extended_name == nm.XSD_ANY_TYPE:
                    self._set_type(self.maps.any_type)
                else:
                    try:
                        self._set_type(self.maps.types[extended_name])
                    except KeyError:
                        self.parse_error(_('unknown type {!r}').format(type_name))
                        self._set_type(self.maps.any_type)
            finally:
                child = self._parse_child_component(self.elem, strict=False)
                if child is not None and child.tag in nm.GLOBAL_TYPES_TAGS:
                    msg = _("the attribute 'type' and a xs:{} local "
                            "declaration are mutually exclusive")
                    self.parse_error(msg.format(child.tag.split('}')[-1]))
        elif (child := self._parse_child_component(self.elem, strict=False)) is None:
            self._set_type(self.maps.any_type)
        else:
            try:
                self._set_type(
                    self.builders.local_types[child.tag](child, self.schema, self)
                )
            except KeyError:
                self._set_type(self.maps.any_type)

    def _parse_constraints(self) -> None:
        # Value constraints
        if 'default' in self.elem.attrib:
            self.default = self.elem.attrib['default']
            if 'fixed' in self.elem.attrib:
                msg = _("'default' and 'fixed' attributes are mutually exclusive")
                self.parse_error(msg)

            if not self.type.text_is_valid(self.default):
                msg = _("'default' value {!r} is not compatible with element's type")
                self.parse_error(msg.format(self.default))
                self.default = None
            elif self.xsd_version == '1.0' and self.type.is_key():
                msg = _("xs:ID or a type derived from xs:ID cannot have a default value")
                self.parse_error(msg)

        elif 'fixed' in self.elem.attrib:
            self.fixed = self.elem.attrib['fixed']
            if not self.type.text_is_valid(self.fixed):
                msg = _("'fixed' value {!r} is not compatible with element's type")
                self.parse_error(msg.format(self.fixed))
                self.fixed = None
            elif self.xsd_version == '1.0' and self.type.is_key():
                msg = _("xs:ID or a type derived from xs:ID cannot have a fixed value")
                self.parse_error(msg)

        # Identity constraints
        self.identities = []
        for child in self.elem:
            if child.tag in nm.IDENTITY_TAGS:
                identity = self.builders.identities[child.tag](child, self.schema, self)
                if identity.ref:
                    if any(identity.name == x.name for x in self.identities):
                        msg = _("duplicated identity constraint %r:")
                        self.parse_error(msg % identity.name, child)

                    self.identities.append(identity)
                    continue

                try:
                    if child != self.maps.identities[identity.name].elem:
                        msg = _("duplicated identity constraint %r:")
                        self.parse_error(msg % identity.name, child)
                except KeyError:
                    self.maps.identities[identity.name] = identity
                finally:
                    self.identities.append(identity)

    def _parse_substitution_group(self, substitution_group: str) -> None:
        try:
            substitution_group_qname = self.schema.resolve_qname(substitution_group)
        except (KeyError, ValueError, RuntimeError) as err:
            self.parse_error(err)
            return
        else:
            if substitution_group_qname[0] != '{':
                substitution_group_qname = get_qname(
                    self.target_namespace, substitution_group_qname
                )

        try:
            head_element = self.maps.elements[substitution_group_qname]
        except KeyError:
            msg = _("unknown substitutionGroup %r")
            self.parse_error(msg % substitution_group)
            return
        else:
            if isinstance(head_element, tuple):
                msg = _("circularity found for substitutionGroup %r")
                self.parse_error(msg % substitution_group)
                return
            elif 'substitution' in head_element.block:
                return

        final = head_element.final
        if self.type.name == nm.XSD_ANY_TYPE and 'type' not in self.elem.attrib:
            if head_element.type.name != nm.XSD_ANY_TYPE:
                # Set the type with head element's type for validate content
                # ref: https://www.w3.org/TR/xmlschema-1/#cElement_Declarations
                self.type = head_element.type
        elif not self.type.is_derived(head_element.type):
            msg = _("{0!r} type is not of the same or a derivation "
                    "of the head element {1!r} type")
            self.parse_error(msg.format(self, head_element))
        elif final == '#all' or 'extension' in final and 'restriction' in final:
            msg = _("head element %r can't be substituted by an "
                    "element that has a derivation of its type")
            self.parse_error(msg % head_element)
        elif 'extension' in final and self.type.is_derived(head_element.type, 'extension'):
            msg = _("head element %r can't be substituted by an "
                    "element that has an extension of its type")
            self.parse_error(msg % head_element)
        elif 'restriction' in final and self.type.is_derived(head_element.type, 'restriction'):
            msg = _("head element %r can't be substituted by an "
                    "element that has a restriction of its type")
            self.parse_error(msg % head_element)

        try:
            self.maps.substitution_groups[substitution_group_qname].add(self)
        except KeyError:
            self.maps.substitution_groups[substitution_group_qname] = {self}
        finally:
            self.substitution_group = substitution_group_qname

    @property
    def xpath_proxy(self) -> XMLSchemaProxy:
        return XMLSchemaProxy(self.schema, self)

    # noinspection PyTypeChecker
    @property
    def xpath_node(self) -> SchemaElementNode:
        schema_node = self.schema.xpath_node
        node = schema_node.get_element_node(self)
        if isinstance(node, SchemaElementNode):
            return node

        return build_schema_node_tree(
            root=self,
            elements=schema_node.elements,
            global_elements=schema_node.children,
        )

    @property
    def scope(self) -> str:
        """The scope of the element declaration that can be 'global' or 'local'."""
        return 'global' if self.parent is None else 'local'

    @property
    def value_constraint(self) -> Optional[str]:
        """The fixed or the default value if either is defined, `None` otherwise."""
        return self.fixed if self.fixed is not None else self.default

    @property
    def final(self) -> str:
        """
        The effective value for prevent the usage of derived elements. Can be empty, '#all'
        or containing a subset of words (extension|restrictions) separated by a space.
        """
        if self.ref is not None:
            return self.ref.final
        elif self._final is not None:
            return self._final
        return self.schema.final_default

    @property
    def block(self) -> str:
        """
        The effective value for blocking the derivation of the element. Can be empty, '#all' or
        containing a subset of words (extension|restrictions|substitution) separated by a space.
        """
        if self.ref is not None:
            return self.ref.block
        elif self._block is not None:
            return self._block
        return self.schema.block_default

    def overall_min_occurs(self, particle: ModelParticleType) -> int:
        """
        Returns the overall minimum for occurrences of a content model particle.
        The content type of the element must be 'element-only' or 'mixed'.
        """
        return self.type.overall_min_occurs(particle)

    def overall_max_occurs(self, particle: ModelParticleType) -> Optional[int]:
        """
        Returns the overall maximum for occurrences of a content model particle.
        The content type of the element must be 'element-only' or 'mixed'.
        """
        return self.type.overall_max_occurs(particle)

    def get_binding(self, *bases: type[Any], replace_existing: bool = False, **attrs: Any) \
            -> DataBindingType:
        """
        Gets data object binding for XSD element, creating a new one if it doesn't exist.

        :param bases: base classes to use for creating the binding class.
        :param replace_existing: provide `True` to replace an existing binding class.
        :param attrs: attribute and method definitions for the binding class body.
        """
        if self.binding is None or replace_existing:
            if not bases:
                bases = (dataobjects.DataElement,)
            attrs['xsd_element'] = self
            class_name = '{}Binding'.format(self.local_name.title().replace('_', ''))
            self.binding = dataobjects.DataBindingMeta(class_name, bases, attrs)

        return self.binding

    def get_alternative_type(self, elem: Union[ElementType, ElementData],
                             inherited: Optional[dict[str, Any]] = None) -> BaseXsdType:
        return self.type

    def get_attributes(self, xsd_type: BaseXsdType) -> 'XsdAttributeGroup':
        if not isinstance(xsd_type, XsdSimpleType):
            return xsd_type.attributes
        elif xsd_type is self.type:
            return self.attributes
        else:
            return self.builders.create_empty_attribute_group(self)

    def get_path(self, ancestor: Optional[XsdComponent] = None,
                 reverse: bool = False) -> Optional[str]:
        """
        Returns the XPath expression of the element. The path is relative to the schema instance
        in which the element is contained or is relative to a specific ancestor passed as argument.
        In the latter case returns `None` if the argument is not an ancestor.

        :param ancestor: optional XSD component of the same schema, that maybe \
        an ancestor of the element.
        :param reverse: if set to `True` returns the reverse path, from the element to ancestor.
        """
        path: list[str] = []
        xsd_component: Optional[XsdComponent] = self
        while xsd_component is not None:
            if xsd_component is ancestor:
                return '/'.join(reversed(path)) or '.'
            elif isinstance(xsd_component, XsdElement):
                path.append('..' if reverse else xsd_component.name)
            xsd_component = xsd_component.parent
        else:
            if ancestor is None:
                return '/'.join(reversed(path)) or '.'
            return None

    def iter_components(self, xsd_classes: Optional[ComponentClassType] = None) \
            -> Iterator[XsdComponent]:

        if xsd_classes is None:
            yield self
            yield from self.identities
        else:
            if isinstance(self, xsd_classes):
                yield self
            if issubclass(XsdIdentity, xsd_classes):
                yield from self.identities

        if self.ref is None and self.type.parent is not None:
            yield from self.type.iter_components(xsd_classes)

    def iter_substitutes(self) -> Iterator['XsdElement']:
        if self.parent is None or self.ref is not None:
            if substitutes := self.maps.substitution_groups.get(self.name):
                for xsd_element in substitutes:
                    if not xsd_element.abstract:
                        yield xsd_element
                    for e in xsd_element.iter_substitutes():
                        if not e.abstract:
                            yield e

    def data_value(self, elem: ElementType) -> DecodedValueType:
        """Returns the decoded data value of the provided element as XPath fn:data()."""
        text = elem.text
        if text is None:
            text = self.fixed if self.fixed is not None else self.default
            if text is None:
                return '' if self.type.text_is_valid('') else None
        return self.type.text_decode(text)

    def check_dynamic_context(self, elem: ElementType, validation: str,
                              context: ValidationContext) -> None:
        for ns, url in iter_schema_location_hints(elem):
            if self.maps.get_schema(ns, url, context.source.base_url) is not None:
                continue

            if ns in iter_schema_namespaces(context.source.root, elem):
                reason = _("schemaLocation declaration after namespace start")
                context.validation_error(validation, self, reason, elem)

            try:
                with self.maps.protect_status():
                    if ns in self.maps.namespaces:
                        schema = self.maps.namespaces[ns][0]
                        schema.include_schema(url, context.source.base_url)
                    else:
                        schema = self.schema
                        schema.import_schema(ns, url, context.source.base_url)
                    schema.clear()
                    schema.build()

            except (XMLSchemaValidationError, XMLResourceParseError) as err:
                context.validation_error(validation, self, err, elem)
            except XMLSchemaParseError as err:
                context.validation_error(validation, self, err.message, elem)
            except OSError:
                continue

    def raw_decode(self, obj: ElementType, validation: str, context: ValidationContext) -> Any:
        """
        Decode an Element instance.

        :param obj: the Element that has to be decoded.
        :param validation: the validation mode. Can be 'lax', 'strict' or 'skip'.
        :param context: the decoding context.
        :return: a decoded object.
        """
        error: Union[XMLSchemaValueError, XMLSchemaValidationError]
        result: Any

        if self.abstract:
            if self.name == obj.tag:
                reason = _("can't use an abstract element in an instance")
                context.validation_error(validation, self, reason, obj)
            elif self.name not in self.maps.substitution_groups:
                reason = _("can't use an abstract XSD element for validation "
                           "unless it's the head of a substitution group")
                context.validation_error(validation, self, reason, obj)
            else:
                for xsd_element in self.iter_substitutes():
                    if obj.tag == xsd_element.name:
                        return xsd_element.raw_decode(obj, validation, context)
                else:
                    reason = _("can't use an abstract XSD element for validation")
                    context.validation_error(validation, self, reason, obj)

        if context.validation_hook is not None:
            # Control validation on element and its descendants or stop validation
            _validation = context.validation_hook(obj, self)
            if _validation:
                if isinstance(_validation, str) and _validation in XSD_VALIDATION_MODES:
                    context = _copy(context)
                    validation = _validation
                else:
                    return Empty

        context.elem = obj

        for identity in self.identities:
            if identity in context.identities:
                context.identities[identity].reset(obj)
            else:
                context.identities[identity] = identity.get_counter(obj)

        if not context.level:
            # Need to set converter context with the right object (the resource can be lazy)
            context.converter.set_xmlns_context(obj, context.level)
        elif context.use_location_hints:
            # Use location hints for dynamic schema load
            self.check_dynamic_context(obj, validation, context)

        inherited = context.inherited
        value = content = None
        nilled = False

        if not self.alternatives:
            xsd_type = self.type
        else:
            xsd_type = self.get_alternative_type(obj, inherited)

        if nm.XSI_TYPE in obj.attrib and self.schema.meta_schema is not None:
            # Meta-schema elements ignore xsi:type (issue #350)
            type_name = obj.attrib[nm.XSI_TYPE].strip()
            try:
                xsd_type = self.maps.get_instance_type(
                    type_name, xsd_type, context.namespaces
                )
            except (KeyError, TypeError) as err:
                context.validation_error(validation, self, err, obj)
            else:
                if xsd_type.is_blocked(self):
                    reason = _("usage of %r is blocked") % xsd_type
                    context.validation_error(validation, self, reason, obj)
                elif xsd_type not in self.xsi_types:
                    self.xsi_types.add(xsd_type)

                    # For complex contents augments permanently the XSD elements
                    # that collect keys/keyrefs for enabled identities.
                    if xsd_type.has_complex_content():
                        xpath_element = XPathElement(self.name, xsd_type)
                        for counter in context.identities.values():
                            if counter.enabled:
                                try:
                                    counter.identity.update_elements(xpath_element)
                                except TypeError as e:
                                    context.validation_error(validation, self, e, obj)

        if xsd_type.abstract:
            reason = _("%r is abstract") % xsd_type
            context.validation_error(validation, self, reason, obj)

        id_list = context.id_list
        if xsd_type.is_complex() and self.xsd_version == '1.1':
            # Track XSD 1.1 multiple xs:ID attributes/children
            context.id_list = []

        content_decoder = xsd_type if isinstance(xsd_type, XsdSimpleType) else xsd_type.content

        # Decode attributes
        attribute_group = self.get_attributes(xsd_type)
        context.level += 1
        attributes = attribute_group.raw_decode(obj.attrib, validation, context)
        context.level -= 1

        if self.inheritable and any(name in self.inheritable for name in obj.attrib):
            if inherited:
                inherited = inherited.copy()
                inherited.update((k, v) for k, v in obj.attrib.items() if k in self.inheritable)
            else:
                inherited = {k: v for k, v in obj.attrib.items() if k in self.inheritable}
            context = _copy(context)
            context.inherited = inherited

        # Checks the xsi:nil attribute of the instance
        if nm.XSI_NIL in obj.attrib:
            xsi_nil = obj.attrib[nm.XSI_NIL].strip()
            if not self.nillable:
                reason = _("element is not nillable")
                context.validation_error(validation, self, reason, obj)
            elif xsi_nil not in ('0', '1', 'false', 'true'):
                reason = _("xsi:nil attribute must have a boolean value")
                context.validation_error(validation, self, reason, obj)
            elif xsi_nil in ('0', 'false'):
                pass
            elif self.fixed is not None:
                reason = _("xsi:nil='true' but the element has a fixed value")
                context.validation_error(validation, self, reason, obj)
            elif obj.text is not None or len(obj):
                reason = _("xsi:nil='true' but the element is not empty")
                context.validation_error(validation, self, reason, obj)
            else:
                nilled = True

        if xsd_type.is_empty() and obj.text and xsd_type.normalize(obj.text):
            reason = _("character data is not allowed because content is empty")
            context.validation_error(validation, self, reason, obj)

        if nilled:
            pass
        elif not isinstance(content_decoder, XsdSimpleType):
            if not isinstance(xsd_type, XsdSimpleType):
                for assertion in xsd_type.assertions:
                    assertion(obj, validation, context)

            context.level += 1
            content = content_decoder.raw_decode(obj, validation, context)
            context.level -= 1

            if content and len(content) == 1 and content[0][0] == 1:
                value, content = content[0][1], None

            if self.fixed is not None and \
                    (len(obj) > 0 or value is not None and self.fixed != value):
                reason = _("must have the fixed value %r") % self.fixed
                context.validation_error(validation, self, reason, obj)

        else:
            if len(obj):
                reason = _("a simple content element can't have child elements")
                context.validation_error(validation, self, reason, obj)

            text = obj.text
            if self.fixed is not None:
                if not text:
                    text = self.fixed
                elif text == self.fixed:
                    pass
                elif not strictly_equal(xsd_type.text_decode(text, context=context),
                                        xsd_type.text_decode(self.fixed)):
                    reason = _("must have the fixed value %r") % self.fixed
                    context.validation_error(validation, self, reason, obj)

            elif not text and self.default is not None and context.use_defaults:
                text = self.default

            if not isinstance(xsd_type, XsdSimpleType):
                for assertion in xsd_type.assertions:
                    assertion(obj, validation, context, value=text)

                if text and content_decoder.is_list():
                    value = text.split()
                else:
                    value = text

            elif xsd_type.is_notation():
                if xsd_type.name == nm.XSD_NOTATION_TYPE:
                    msg = _("cannot validate against xs:NOTATION directly, "
                            "only against a subtype with an enumeration facet")
                    context.validation_error(validation, self, msg, text)
                elif not xsd_type.enumeration:
                    msg = _("missing enumeration facet in xs:NOTATION subtype")
                    context.validation_error(validation, self, msg, text)

            result = content_decoder.raw_decode(text or '', validation, context)
            if not isinstance(context, DecodeContext):
                value = result
            else:
                if result is None and context.filler is not None:
                    value = context.filler(self)
                elif text or context.keep_empty:
                    value = result

                if context.value_hook is not None:
                    value = context.value_hook(value, xsd_type)
                elif isinstance(value, context.keep_datatypes) or value is None:
                    pass
                elif isinstance(value, str):
                    if value[:1] == '{' and xsd_type.is_qname():
                        value = text
                elif isinstance(value, Decimal):
                    if context.decimal_type is not None:
                        value = context.decimal_type(value)
                elif isinstance(value, (AbstractDateTime, Duration)):
                    value = str(value) if text is None else text.strip()
                else:
                    value = str(value)

        context.id_list = id_list
        xmlns = context.converter.set_xmlns_context(obj, context.level)  # Purge sub-contexts

        if isinstance(context, DecodeContext):
            element_data = ElementData(obj.tag, value, content, attributes, xmlns)
            if context.element_hook is not None:
                element_data = context.element_hook(element_data, self, xsd_type)

            try:
                result = context.converter.element_decode(
                    element_data, self, xsd_type, context.level
                )
            except (ValueError, TypeError) as err:
                context.validation_error(validation, self, err, obj)
                result = None
        elif not context.level:
            result = ElementData(obj.tag, value, None, attributes, None)
        else:
            result = None

        if content is not None:
            del content

        if self.selected_by:
            self.collect_key_fields(obj, xsd_type, validation, nilled, context)

        # Apply non XSD optional validations
        if context.extra_validator is not None:
            try:
                errors = context.extra_validator(obj, self)
            except XMLSchemaValidationError as err:
                context.validation_error(validation, self, err, obj)
            else:
                if isinstance(errors, GeneratorType):
                    for error in errors:
                        context.validation_error(validation, self, error, obj)

        # Disable collect for out of scope identities and check key references
        if context.max_depth is None:
            for identity in self.identities:
                counter = context.identities[identity]
                counter.enabled = False
                if isinstance(identity, XsdKeyref):
                    assert isinstance(counter, KeyrefCounter)
                    for error in counter.iter_errors(context.identities):
                        context.validation_error(validation, self, error, obj)
        elif context.level:
            for identity in self.identities:
                context.identities[identity].enabled = False

        return result

    def collect_key_fields(self, obj: ElementType, xsd_type: BaseXsdType,
                           validation: str, nilled: bool, context: ValidationContext) -> None:

        element_node: Union[EtreeElementNode, LazyElementNode]
        element_node = cast(EtreeElementNode, context.source.get_xpath_node(obj))

        xsd_element = self if self.ref is None else self.ref
        if xsd_element.type is not xsd_type:
            xsd_element = _copy(xsd_element)
            xsd_element._set_type(xsd_type)

        # Collect field values for identities that refer to this XSD element.
        for identity in self.selected_by:
            try:
                counter = context.identities[identity]
            except KeyError:
                continue
            else:
                if not counter.enabled:
                    continue

            if counter.elements is None:
                # Apply selector on Element ancestor for obtain the selected elements
                root_node = context.source.get_xpath_node(counter.elem)
                xpath_context = XPathContext(root_node)
                assert identity.selector is not None
                counter.elements = {
                    x for x in identity.selector.token.select_results(xpath_context)
                }

            if obj not in counter.elements:
                continue

            if xsd_element in identity.elements:
                selectors = identity.elements[xsd_element]
            else:
                # noinspection PyTypeChecker
                selectors = [FieldValueSelector(f, xsd_element) for f in identity.fields]

            try:
                fields = tuple(
                    s.get_value(element_node, context.namespaces) for s in selectors
                )
            except (XMLSchemaValueError, XMLSchemaTypeError) as err:
                context.validation_error(validation, self, err, obj)
            else:
                if any(x is not None for x in fields) or nilled:
                    try:
                        counter.increase(fields)
                    except ValueError as err:
                        context.validation_error(validation, self, err, obj)

    def to_objects(self, obj: ElementType, with_bindings: bool = False, **kwargs: Any) \
            -> DecodeType['dataobjects.DataElement']:
        """
        Decodes XML data to Python data objects.

        :param obj: the XML data source.
        :param with_bindings: if `True` is provided the decoding is done using \
        :class:`DataBindingConverter` that used XML data binding classes. For \
        default the objects are instances of :class:`DataElement` and uses the \
        :class:`DataElementConverter`.
        :param kwargs: other optional keyword arguments for the method \
        :func:`iter_decode`, except the argument *converter*.
        """
        if with_bindings:
            return self.decode(obj, converter=dataobjects.DataBindingConverter, **kwargs)
        return self.decode(obj, converter=dataobjects.DataElementConverter, **kwargs)

    def raw_encode(self, obj: Any, validation: str, context: EncodeContext) \
            -> Optional[ElementType]:
        """
        Encode data to an Element.

        :param obj: the data that has to be encoded.
        :param validation: the validation mode. Can be 'lax', 'strict' or 'skip'.
        :param context: the encoding context.
        :return: returns an Element.
        """
        errors: list[Union[str, Exception]] = []

        try:
            element_data = context.converter.element_encode(obj, self, context.level)
        except (ValueError, TypeError) as err:
            context.validation_error(validation, self, err, obj)
            return None

        if context.max_depth is not None and context.max_depth == 0 and not context.level:
            return None

        tag, text, content, attributes, xmlns = element_data
        context.elem = elem = context.create_element(tag)

        if self.abstract:
            if self.name == tag and context.converter.losslessly:
                reason = _("can't use an abstract element in an instance")
                context.validation_error(validation, self, reason, obj)
            elif self.name not in self.maps.substitution_groups:
                reason = _("can't use an abstract XSD element for validation "
                           "unless it's the head of a substitution group")
                context.validation_error(validation, self, reason, obj)
            else:
                for xsd_element in self.iter_substitutes():
                    if tag == xsd_element.name:
                        return xsd_element.raw_encode(obj, validation, context)
                else:
                    # In some cases the original tag could be missed, so try each
                    # substitute before generate an error.
                    for xsd_element in self.iter_substitutes():
                        return xsd_element.raw_encode(obj, validation, context)
                    else:
                        reason = _("can't use an abstract XSD element for validation")
                        context.validation_error(validation, self, reason, obj)

        if not self.alternatives:
            xsd_type = self.type
        else:
            xsd_type = self.get_alternative_type(element_data)

        if nm.XSI_TYPE in attributes and self.schema.meta_schema is not None:
            type_name = attributes[nm.XSI_TYPE].strip()
            try:
                xsd_type = self.maps.get_instance_type(
                    type_name, xsd_type, context.namespaces
                )
            except (KeyError, TypeError) as err:
                errors.append(err)
            else:
                default_namespace = context.converter.get('')
                if default_namespace and not isinstance(xsd_type, XsdSimpleType):
                    # Adjust attributes mapped into default namespace

                    ns_part = f'{{{default_namespace}}}'
                    for k in list(attributes):
                        if not k.startswith(ns_part):
                            continue
                        elif k in xsd_type.attributes:
                            continue

                        local_name = k[len(ns_part):]
                        if local_name in xsd_type.attributes:
                            attributes[local_name] = attributes[k]
                            del attributes[k]

        attribute_group = self.get_attributes(xsd_type)
        context.level += 1
        try:
            elem.attrib.update(attribute_group.raw_encode(attributes, validation, context))
        except XMLSchemaValidationError:
            elem.attrib.update(raw_encode_attributes(attributes))
            raise

        context.level -= 1

        if nm.XSI_NIL in attributes:
            xsi_nil = attributes[nm.XSI_NIL].strip()
            if not self.nillable:
                errors.append("element is not nillable.")
            elif xsi_nil not in ('0', '1', 'true', 'false'):
                errors.append("xsi:nil attribute must has a boolean value.")
            elif xsi_nil in ('0', 'false'):
                pass
            elif self.fixed is not None:
                errors.append("xsi:nil='true' but the element has a fixed value.")
            elif text not in (None, '') or content:
                errors.append("xsi:nil='true' but the element is not empty.")
            else:
                for e in errors:
                    context.validation_error(validation, self, e, elem)
                return elem

        if isinstance(xsd_type, XsdSimpleType):
            if content:
                errors.append("a simpleType element can't have child elements.")

            if text is not None:
                try:
                    elem.text = xsd_type.raw_encode(text, validation, context)
                except XMLSchemaValidationError as err:
                    if err.elem is not None:
                        raise
                    errors.append(err)

            elif self.fixed is not None:
                elem.text = self.fixed
            elif self.default is not None and context.use_defaults:
                elem.text = self.default

        elif isinstance(xsd_type.content, XsdSimpleType):
            if xsd_type.content.max_length == 0:
                pass
            elif text is not None:
                try:
                    elem.text = xsd_type.content.raw_encode(text, validation, context)
                except XMLSchemaValidationError as err:
                    if err.elem is not None:
                        raise
                    errors.append(err)

            elif self.fixed is not None:
                elem.text = self.fixed
            elif self.default is not None and context.use_defaults:
                elem.text = self.default

        else:
            context.level += 1
            xsd_type.content.raw_encode(element_data, validation, context)
            context.level -= 1

        if errors:
            for e in errors:
                context.validation_error(validation, self, e, elem)

        del element_data
        return elem

    def is_matching(self, name: Optional[str], default_namespace: Optional[str] = None,
                    group: Optional['XsdGroup'] = None, **kwargs: Any) -> bool:
        if not name:
            return False
        elif default_namespace and name[0] != '{':
            name = f'{{{default_namespace}}}{name}'

            # Workaround for backward compatibility of XPath selectors on schemas.
            if not self.qualified and default_namespace == self.target_namespace:
                return (name == self.qualified_name or
                        any(name == e.qualified_name for e in self.iter_substitutes()))

        return name == self.name or name in self.substitutes

    def match(self, name: Optional[str], default_namespace: Optional[str] = None,
              **kwargs: Any) -> Optional['XsdElement']:
        if not name:
            return None
        elif default_namespace and name[0] != '{':
            name = f'{{{default_namespace}}}{name}'

        if name == self.name:
            return self
        elif name in self.substitutes:
            return self.maps.elements[name]
        return None

    @schema_cache
    def match_child(self, name: str) -> Optional['XsdElement']:
        xsd_group = self.type.model_group
        if xsd_group is None:
            # fallback to xs:anyType encoder for matching extra content
            xsd_group = self.maps.any_type.model_group
            assert xsd_group is not None

        for xsd_child in xsd_group.elements:
            matched_element = xsd_child.match(name, resolve=True)
            if isinstance(matched_element, XsdElement):
                return matched_element
        else:
            if name in self.maps.elements and xsd_group.open_content_mode != 'none':
                return self.maps.elements[name]
            return None

    @schema_cache
    def is_restriction(self, other: ModelParticleType, check_occurs: bool = True) -> bool:
        e: ModelParticleType

        if isinstance(other, XsdAnyElement):
            if self.min_occurs == self.max_occurs == 0:
                return True
            if check_occurs and not self.has_occurs_restriction(other):
                return False
            return other.is_matching(self.name, self.default_namespace)
        elif isinstance(other, XsdElement):
            if self.name != other.name:
                if other.name == self.substitution_group and \
                        other.min_occurs != other.max_occurs and \
                        self.max_occurs != 0 and not other.abstract \
                        and self.xsd_version == '1.0':
                    # A UPA violation case. Base is the head element, it's not
                    # abstract and has non-deterministic occurs: this is less
                    # restrictive than W3C test group (elemZ026), marked as
                    # invalid despite it's based on an abstract declaration.
                    # See also test case invalid_restrictions1.xsd.
                    return False

                for e in other.iter_substitutes():
                    if e.name == self.name:
                        break
                else:
                    return False

            if check_occurs and not self.has_occurs_restriction(other):
                return False
            elif self.max_occurs == 0 and check_occurs:
                return True  # type is not effective if the element can't have occurrences
            elif not self.is_consistent(other) and self.type.elem is not other.type.elem and \
                    not self.type.is_derived(other.type, 'restriction') and not other.type.abstract:
                return False
            elif other.fixed is not None and \
                    (self.fixed is None or self.type.normalize(
                        self.fixed) != other.type.normalize(other.fixed)):
                return False
            elif other.nillable is False and self.nillable:
                return False
            elif any(value not in self.block for value in other.block.split()):
                return False
            elif not all(k in other.identities for k in self.identities):
                return False
            else:
                return True
        elif other.model == 'choice':
            if other.is_empty() and self.max_occurs != 0:
                return False

            check_group_items_occurs = self.xsd_version == '1.0'
            total_occurs = OccursCalculator()
            for e in other.iter_model():
                if not isinstance(e, (XsdElement, XsdAnyElement)):
                    return False
                elif not self.is_restriction(e, check_group_items_occurs):
                    continue
                total_occurs += e
                total_occurs *= other
                if self.has_occurs_restriction(total_occurs):
                    return True
                total_occurs.reset()
            return False
        else:
            match_restriction = False
            for e in other.iter_model():
                if match_restriction:
                    if not e.is_emptiable():
                        return False
                elif self.is_restriction(e):
                    match_restriction = True
                elif not e.is_emptiable():
                    return False
            return True

    @schema_cache
    def is_overlap(self, other: SchemaElementType) -> bool:
        if isinstance(other, XsdElement):
            if self.name == other.name:
                return True
            elif other.substitution_group == self.name or other.name == self.substitution_group:
                return True
        elif isinstance(other, XsdAnyElement):
            if other.is_matching(self.name, self.default_namespace):
                return True
            for e in self.maps.substitution_groups.get(self.name, ()):
                if other.is_matching(e.name, self.default_namespace):
                    return True
        return False

    def is_consistent(self, other: SchemaElementType, strict: bool = True) -> bool:
        """
        Element Declarations Consistent check between two element particles.

        Ref: https://www.w3.org/TR/xmlschema-1/#cos-element-consistent

        :returns: `True` if there is no inconsistency between the particles, `False` otherwise,
        """
        return self.name != other.name or self.type is other.type

    def is_single(self) -> bool:
        if self.parent is None:
            return True
        elif self.max_occurs != 1:
            return False
        elif self.parent.max_occurs == 1:
            return True
        else:
            return self.parent.model != 'choice' and len(self.parent) > 1

    def is_substitute(self, other: ModelParticleType) -> bool:
        return not self.abstract and isinstance(other, XsdElement) \
            and self.substitution_group == other.name


class Xsd11Element(XsdElement):
    """
    Class for XSD 1.1 *element* declarations.

    ..  <element
          abstract = boolean : false
          block = (#all | List of (extension | restriction | substitution))
          default = string
          final = (#all | List of (extension | restriction))
          fixed = string
          form = (qualified | unqualified)
          id = ID
          maxOccurs = (nonNegativeInteger | unbounded)  : 1
          minOccurs = nonNegativeInteger : 1
          name = NCName
          nillable = boolean : false
          ref = QName
          substitutionGroup = List of QName
          targetNamespace = anyURI
          type = QName
          {any attributes with non-schema namespace . . .}>
          Content: (annotation?, ((simpleType | complexType)?, alternative*,
          (unique | key | keyref)*))
        </element>
    """
    def _parse(self) -> None:
        if self._built is not None and isinstance(self.parent, MutableSequence):
            return

        self.min_occurs = self.max_occurs = 1
        self.selected_by = set()
        self.xsi_types = set()
        self._parse_particle(self.elem)
        self._parse_attributes()

        if self.ref is None:
            self._parse_type()
            self._parse_alternatives()
            self._parse_constraints()

            if self.parent is None:
                self.substitutes = set()
                if 'substitutionGroup' in self.elem.attrib:
                    for substitution_group in self.elem.attrib['substitutionGroup'].split():
                        self._parse_substitution_group(substitution_group)

        if 'targetNamespace' in self.elem.attrib:
            self._parse_target_namespace()

        if any(v.inheritable for v in self.attributes.values()):
            self.inheritable = {}
            for k, v in self.attributes.items():
                if k is not None and isinstance(v, XsdAttribute):
                    if v.inheritable:
                        self.inheritable[k] = v

        self._built = True

    def _parse_alternatives(self) -> None:
        alternatives = []
        has_test = True
        for child in self.elem:
            if child.tag == nm.XSD_ALTERNATIVE:
                alternatives.append(XsdAlternative(child, self.schema, self))
                if not has_test:
                    msg = _("test attribute missing in non-final alternative")
                    self.parse_error(msg)
                has_test = 'test' in child.attrib

        if alternatives:
            self.alternatives = alternatives

    def iter_components(self, xsd_classes: ComponentClassType = None) -> Iterator[XsdComponent]:
        if xsd_classes is None:
            yield self
            yield from self.identities
        else:
            if isinstance(self, xsd_classes):
                yield self
            if issubclass(XsdIdentity, xsd_classes):
                yield from self.identities

        for alt in self.alternatives:
            yield from alt.iter_components(xsd_classes)

        if self.ref is None and self.type.parent is not None:
            yield from self.type.iter_components(xsd_classes)

    def iter_substitutes(self) -> Iterator[XsdElement]:
        if self.parent is None or self.ref is not None:
            if substitutes := self.maps.substitution_groups.get(self.name):
                for xsd_element in substitutes:
                    yield xsd_element
                    yield from xsd_element.iter_substitutes()

    def get_alternative_type(self, elem: Union[ElementType, ElementData],
                             inherited: Optional[dict[str, Any]] = None) -> BaseXsdType:
        if isinstance(elem, ElementData):
            if elem.attributes:
                attrib = raw_encode_attributes(elem.attributes)
                elem = Element(elem.tag, attrib=attrib)
            else:
                elem = Element(elem.tag)

        if inherited:
            dummy = Element('_dummy_element', attrib=inherited)
            dummy.attrib.update(elem.attrib)

            for alt in self.alternatives:
                if alt.type is not None:
                    if alt.token is None or alt.test(elem) or alt.test(dummy):
                        return alt.type
        else:
            for alt in self.alternatives:
                if alt.type is not None:
                    if alt.token is None or alt.test(elem):
                        return alt.type

        return self.type

    def is_overlap(self, other: SchemaElementType) -> bool:
        if isinstance(other, XsdElement):
            if self.name == other.name:
                return True
            elif any(self.name == x.name for x in other.iter_substitutes()):
                return True

            for e in self.iter_substitutes():
                if other.name == e.name or any(x is e for x in other.iter_substitutes()):
                    return True

        elif isinstance(other, XsdAnyElement):
            if other.is_matching(self.name, self.default_namespace):
                return True
            for e in self.maps.substitution_groups.get(self.name, ()):
                if other.is_matching(e.name, self.default_namespace):
                    return True
        return False

    def is_consistent(self, other: SchemaElementType, strict: bool = True) -> bool:
        if isinstance(other, XsdAnyElement):
            if other.process_contents == 'skip':
                return True
            xsd_element = other.match(self.name, self.default_namespace, resolve=True)
            return xsd_element is None or self.is_consistent(xsd_element, strict=False)

        e1: XsdElement = self
        e2 = other
        if self.name != other.name:
            for e1 in self.iter_substitutes():
                if e1.name == other.name:
                    break
            else:
                for e2 in other.iter_substitutes():
                    if e2.name == self.name:
                        break
                else:
                    return True

        if len(e1.alternatives) != len(e2.alternatives):
            return False
        elif e1.type is not e2.type and strict:
            return False
        elif e1.type is not e2.type or \
                not all(any(a == x for x in e2.alternatives) for a in e1.alternatives) or \
                not all(any(a == x for x in e1.alternatives) for a in e2.alternatives):
            msg = _("Maybe a not equivalent type table between elements {0!r} and {1!r}")
            warnings.warn(msg.format(e1, e2), XMLSchemaTypeTableWarning, stacklevel=3)
        return True

    def check_dynamic_context(self, elem: ElementType, validation: str,
                              context: ValidationContext) -> None:
        for ns, url in iter_schema_location_hints(elem):
            if self.maps.get_schema(ns, url, context.source.base_url) is not None:
                continue

            try:
                with self.maps.protect_status():
                    if ns in self.maps.namespaces:
                        schema = self.maps.namespaces[ns][0]
                        schema.include_schema(url, context.source.base_url)
                    else:
                        schema = self.schema
                        schema.import_schema(ns, url, context.source.base_url)
                    schema.clear()
                    schema.build()

            except (XMLSchemaValidationError, ParseError) as err:
                context.validation_error(validation, self, err, elem)
            except XMLSchemaParseError as err:
                context.validation_error(validation, self, err.message, elem)
            except OSError:
                continue
            else:
                def stop_validation(e: ElementType, _xsd_element: XsdElement) -> bool:
                    if e is elem:
                        raise XMLSchemaStopValidation()
                    return False

                errors = list(schema.iter_errors(context.source, validation_hook=stop_validation))
                if len(context.errors) != len(errors) or \
                        any(e1.elem is not e2.elem for e1, e2 in zip(context.errors, errors)):
                    reason = _(f"adding schema at {url} change the "
                               f"assessment outcome of previous items")
                    context.validation_error(validation, self, reason, elem)


class XsdAlternative(XsdComponent):
    """
    XSD 1.1 type *alternative* definitions.

    ..  <alternative
          id = ID
          test = an XPath expression
          type = QName
          xpathDefaultNamespace = (anyURI | (##defaultNamespace | ##targetNamespace | ##local))
          {any attributes with non-schema namespace . . .}>
          Content: (annotation?, (simpleType | complexType)?)
        </alternative>
    """
    parent: XsdElement
    type: BaseXsdType
    path: Optional[str]
    token: Optional[XPathToken]
    _ADMITTED_TAGS = nm.XSD_ALTERNATIVE,

    __slots__ = ('xpath_default_namespace', 'path', 'token', 'type')

    def __repr__(self) -> str:
        return '%s(type=%r, test=%r)' % (
            self.__class__.__name__, self.elem.get('type'), self.elem.get('test')
        )

    def __eq__(self, other: object) -> bool:
        return isinstance(other, XsdAlternative) and \
            self.path == other.path and self.type is other.type and \
            self.xpath_default_namespace == other.xpath_default_namespace

    def __ne__(self, other: object) -> bool:
        return not isinstance(other, XsdAlternative) or \
            self.path != other.path or self.type is not other.type or \
            self.xpath_default_namespace != other.xpath_default_namespace

    def _parse(self) -> None:
        attrib = self.elem.attrib

        if 'xpathDefaultNamespace' in attrib:
            self.xpath_default_namespace = parse_xpath_default_namespace(self)
        else:
            self.xpath_default_namespace = self.schema.xpath_default_namespace

        parser = XPath2Parser(
            namespaces=self.schema.namespaces,
            strict=False,
            default_namespace=self.xpath_default_namespace
        )

        try:
            self.path = attrib['test']
        except KeyError:
            # an absent test is not an error, it should be the default type
            self.path = self.token = None
        else:
            try:
                self.token = parser.parse(self.path)
            except ElementPathError as err:
                self.parse_error(err)
                self.token = parser.parse('false()')
                self.path = 'false()'

        try:
            type_qname = self.schema.resolve_qname(attrib['type'])
        except (KeyError, ValueError, RuntimeError) as err:
            if 'type' in attrib:
                self.parse_error(err)
                self.type = self.maps.any_type
            elif (child := self._parse_child_component(self.elem, strict=False)) is None:
                self.parse_error(_("missing 'type' attribute"))
                self.type = self.maps.any_type
            else:
                try:
                    self.type = self.builders.local_types[child.tag](
                        child, self.schema, self
                    )
                except KeyError:
                    self.parse_error(_("missing 'type' attribute"))
                    self.type = self.maps.any_type
                else:
                    if not self.type.is_derived(self.parent.type):
                        msg = _("declared type is not derived from {!r}")
                        self.parse_error(msg.format(self.parent.type))
        else:
            try:
                self.type = self.maps.types[type_qname]
            except KeyError:
                self.parse_error(_("unknown type {!r}").format(attrib['type']))
                self.type = self.maps.any_type
            else:
                if self.type.name != nm.XSD_ERROR and not self.type.is_derived(self.parent.type):
                    msg = _("type {0!r} is not derived from {1!r}")
                    self.parse_error(msg.format(attrib['type'], self.parent.type))

                child = self._parse_child_component(self.elem, strict=False)
                if child is not None and child.tag in nm.GLOBAL_TYPES_TAGS:
                    msg = _("the attribute 'type' and the xs:%s local "
                            "declaration are mutually exclusive")
                    self.parse_error(msg % child.tag.split('}')[-1])

    @property
    def validation_attempted(self) -> str:
        if self._built:
            return 'full'
        elif not hasattr(self, 'type'):
            return 'none'
        else:
            return self.type.validation_attempted

    def iter_components(self, xsd_classes: ComponentClassType = None) -> Iterator[XsdComponent]:
        if xsd_classes is None or isinstance(self, xsd_classes):
            yield self
        if self.type is not None and self.type.parent is not None:
            yield from self.type.iter_components(xsd_classes)

    def test(self, elem: ElementType) -> bool:
        if self.token is None:
            return False

        try:
            result = list(self.token.select(context=XPathContext(elem)))
            return self.token.boolean_value(result)
        except (TypeError, ValueError):
            return False



#
# Copyright (c), 2025-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from collections import deque
from collections.abc import Callable, Iterable, Iterator
from functools import cached_property
from typing import cast, Optional, TYPE_CHECKING, Union
from xml.etree.ElementTree import Element

from elementpath import XPath2Parser, XPathToken, ElementNode, XPathContext

from xmlschema.aliases import ElementType, NsmapType
from xmlschema.exceptions import XMLSchemaTypeError

if TYPE_CHECKING:
    from xmlschema.resources import XMLResource

CacheKeyType = Union[
    tuple[str, type['ElementSelector']],
    tuple[Union[str, type['ElementSelector'], Iterable[tuple[str, str]], tuple[str, str]], ...]
]

_selectors_cache: dict[CacheKeyType, 'ElementSelector'] = {}
_dummy_element = Element('dummy')


def is_ncname(s: str) -> bool:
    return s.isalpha() and ':' not in s and all(is_ncname_continuation(c) for c in s[1:])


def is_ncname_continuation(c: str) -> bool:
    return (c.isalnum() or c in '-.\u00B7\u0387\u06DD\u06DE\u203F\u2040'
            or 0x300 <= ord(c) <= 0x36F)


def split_path(path: str, namespaces: Optional[NsmapType] = None,
               extended_names: bool = False) -> deque[str]:
    """
    Splits a path expression to a sequence of chunks that put in evidence path steps,
    predicates and other parts that can be useful for checking some properties of the
    provided path, like the path depth of if the path is composed only by path steps
    and wildcards.

    :param path: the path expression to split.
    :param namespaces: an optional namespace mapping to use on prefixed names.
    :param extended_names: if `True` maps prefixed names to extended form. For \
    default only the default namespace is used, if defined and not empty.
    """
    start = end = 0

    def flush() -> None:
        nonlocal start
        if start < end:
            chunks.append(path[start:end])
            start = end

    def advance(condition: Callable[[str], bool]) -> None:
        nonlocal end
        end += 1
        while condition(path[end]):
            end += 1

    path = path.replace(' ', '').replace('\t', '').replace('./', '')  # path normalization
    chunks: deque[str] = deque([''])  # add an empty element to avoid index errors
    default_namespace = None if not namespaces else namespaces.get('')

    while True:
        try:
            flush()
            if path[end] in '"\'':
                advance(lambda x: x != path[start])
                end += 1
            elif path[end] == '{':
                advance(lambda x: x != '}')
                end += 1
                if path[end].isalpha():
                    advance(is_ncname_continuation)
            elif path[end].isalpha():
                advance(is_ncname_continuation)
                if path[end] == ':':
                    prefix = path[start:end]
                    end += 1
                    if path[end].isalpha():
                        advance(is_ncname_continuation)
                        if extended_names and namespaces and prefix in namespaces:
                            flush()
                            uri = namespaces[prefix]
                            chunks[-1] = f'{{{uri}}}{chunks[-1][len(prefix)+1:]}'

                elif default_namespace and not chunks[-1].endswith('@'):
                    flush()
                    chunks[-1] = f'{{{default_namespace}}}{chunks[-1]}'
            elif path[end] == '/':
                advance(lambda x: x == '/')
            else:
                end += 1

        except IndexError:
            if start < len(path):
                flush()
                if default_namespace and is_ncname(chunks[-1]) and chunks[-2] != '@':
                    chunks[-1] = f'{{{default_namespace}}}{chunks[-1]}'

            chunks.popleft()
            return chunks


class ElementSelector:
    """
    An XPath selector for selecting ElementTree elements. Raises an error
    if the path parse fails or is incompatible with the selector type.

    :param path: the XPath expression.
    :param namespaces: an optional namespace mapping.
    """

    path: str
    """The normalized XPath expression of the path provided by argument."""

    namespaces: Optional[dict[str, str]]
    """The namespaces mapping associated with the XPath expression path."""

    _parser: XPath2Parser
    _token: XPathToken

    @classmethod
    def cached_selector(cls, path: str, namespaces: Optional[NsmapType] = None) \
            -> 'ElementSelector':
        """A builder of ElementSelector instances based on a cache."""
        key: CacheKeyType = (path, cls)
        if namespaces is not None:
            key += tuple(sorted(namespaces.items()))

        try:
            return _selectors_cache[key]
        except KeyError:
            if len(_selectors_cache) > 100:
                _selectors_cache.clear()

            selector = cls(path, namespaces)
            _selectors_cache[key] = selector
            return selector

    def __init__(self, path: str, namespaces: Optional[NsmapType] = None) -> None:
        self.namespaces = None if namespaces is None else {k: v for k, v in namespaces.items()}
        self._parts = split_path(path, namespaces)

        self.path = ''.join(self._parts)

        self._parser = XPath2Parser(namespaces, strict=False)
        self._token = self._parser.parse(self.path)
        self.select(_dummy_element)

    def __repr__(self) -> str:
        return '%s(path=%r, )' % (self.__class__.__name__, self.path)

    @property
    def parts(self) -> list[str]:
        """Return a list with the parts of the parsed path."""
        return list(self._parts)

    @cached_property
    def relative_path(self) -> str:
        """The equivalent path expression relative to root element."""
        parts = self._parts.copy()
        if not parts:
            parts.appendleft('.')
        elif parts[0] == '//':
            parts.appendleft('.')
        elif parts[0] == '/':
            parts.popleft()
            while parts:
                if parts[0].startswith('/'):
                    break
                parts.popleft()
            parts.appendleft('.')
        return ''.join(parts)

    @cached_property
    def select_all(self) -> bool:
        """Returns `True` if the path is composed only by wildcards or path steps."""
        return all(c in ('*', '/', '.') for c in self._parts)

    @cached_property
    def depth(self) -> int:
        """Path depth, 0 means a self axis selector, -1 means an unlimited depth."""
        if not self._parts:
            return 0
        elif '//' in self._parts:
            return -1
        elif self._parts[0] == '/':
            return sum(s == '/' for s in self._parts) - 1
        elif self._parts[0] == '.':
            return sum(s == '/' for s in self._parts)
        else:
            return sum(s == '/' for s in self._parts) + 1

    def select(self, root: Union[ElementType, 'XMLResource']) -> list[ElementType]:
        return list(self.iter_select(root))

    def iter_select(self, root: Union[ElementType, 'XMLResource']) -> Iterator[Element]:
        if hasattr(root, 'xpath_root'):
            context = XPathContext(root.xpath_root)
        else:
            context = XPathContext(root)

        for item in self._token.select(context):
            if not isinstance(item, ElementNode):  # pragma: no cover
                msg = "XPath expressions on XML resources can select only elements"
                raise XMLSchemaTypeError(msg)
            yield cast(ElementType, item.obj)


class ElementPathSelector(ElementSelector):
    """
    An XPath selector that uses `xml.etree.ElementPath.iterfind()` for selecting elements.
    """
    def iter_select(self, root: Union[ElementType, 'XMLResource']) -> Iterator[ElementType]:
        if hasattr(root, 'root'):
            yield from root.root.iterfind(self.relative_path, self.namespaces)
        else:
            yield from root.iterfind(self.relative_path, self.namespaces)



#
# Copyright (c), 2023-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
Optional module for handling XPath 3 parsing on XSD 1.1 assertions.
"""
from typing import Optional

from elementpath import XPathToken, XPathContext
from elementpath.xpath3 import XPath3Parser

__all__ = ['XPath3Parser', 'XsdAssertionXPath3Parser']


class XsdAssertionXPath3Parser(XPath3Parser):
    """Parser for XSD 1.1 assertion facets with XPath 3."""


XsdAssertionXPath3Parser.unregister('last')
XsdAssertionXPath3Parser.unregister('position')


# noinspection PyUnusedLocal
@XsdAssertionXPath3Parser.method(
    XsdAssertionXPath3Parser.function('last', nargs=0)
)
def evaluate_last(self: XPathToken, context: Optional[XPathContext] = None) -> None:
    raise self.missing_context("context item size is undefined")  # pragma: no cover


# noinspection PyUnusedLocal
@XsdAssertionXPath3Parser.method(
    XsdAssertionXPath3Parser.function('position', nargs=0)
)
def evaluate_position(self: XPathToken, context: Optional[XPathContext] = None) -> None:
    raise self.missing_context("context item position is undefined")  # pragma: no cover



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
# mypy: ignore-errors
"""
Subpackage with unittest extensions for xmlschema.

Includes common classes and helpers for building test scripts for xmlschema. The main
part is a test factory for creating test cases from lists of paths to XSD or XML files.
The list of cases can be defined within files named "testfiles". These are text files
that contain a list of relative paths to XSD or XML files, that are used to dinamically
build a set of test classes. Each path is followed by a list of options that defines a
custom setting for each test.
"""
from urllib.request import urlopen
from urllib.error import URLError

from ._helpers import iter_nested_items, etree_elements_assert_equal
from ._test_case_classes import XMLSchemaTestCase, XsdValidatorTestCase
from ._builders import make_schema_test_class, make_validation_test_class
from ._factory import get_test_args, xsd_version_number, defuse_data, \
    get_test_program_args_parser, parse_xmlschema_args, run_xmlschema_tests, \
    get_test_line_args_parser, xmlschema_tests_factory
from ._observers import SchemaObserver, ObservedXMLSchema10, ObservedXMLSchema11


def has_network_access(*locations):
    for url in locations:
        try:
            urlopen(url, timeout=10)
        except (URLError, OSError):
            pass
        else:
            return True
    return False


SKIP_REMOTE_TESTS = not has_network_access('https://github.com/')


__all__ = [
    'XsdValidatorTestCase', 'make_schema_test_class', 'make_validation_test_class',
    'get_test_args', 'xsd_version_number', 'defuse_data', 'get_test_program_args_parser',
    'parse_xmlschema_args', 'run_xmlschema_tests', 'get_test_line_args_parser',
    'xmlschema_tests_factory', 'SchemaObserver', 'ObservedXMLSchema10',
    'ObservedXMLSchema11', 'has_network_access', 'iter_nested_items',
    'etree_elements_assert_equal', 'SKIP_REMOTE_TESTS', 'XMLSchemaTestCase'
]



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import logging
import traceback
from collections.abc import Callable
from functools import wraps
from typing import Any, Optional, TypeVar, Union

from xmlschema.exceptions import XMLSchemaValueError
from xmlschema.translation import gettext as _
from xmlschema.resources import XMLResource

logger = logging.getLogger('xmlschema')

LOG_LEVELS = {'DEBUG', 'INFO', 'WARN', 'WARNING', 'ERROR', 'CRITICAL'}


def set_logging_level(level: Union[str, int]) -> None:
    """set logging level of xmlschema's logger."""
    if isinstance(level, str):
        _level = level.strip().upper()
        if _level not in LOG_LEVELS:
            raise XMLSchemaValueError(
                _("{!r} is not a valid loglevel").format(level)
            )
        logger.setLevel(getattr(logging, _level))
    else:
        logger.setLevel(level)


RT = TypeVar('RT')


def logged(func: Callable[..., RT]) -> Callable[..., RT]:
    """
    A decorator for activating a logging level for a function. The keyword
    argument 'loglevel' is obtained from the keyword arguments and used by the
    wrapper function to set the logging level of the decorated function and
    to restore the original level after the call.
    """
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        loglevel: Optional[Union[int, str]] = kwargs.get('loglevel')
        if loglevel is None:
            return func(*args, **kwargs)
        else:
            current_level = logger.level
            set_logging_level(loglevel)
            try:
                return func(*args, **kwargs)
            finally:
                logger.setLevel(current_level)

    return wrapper


def format_xmlschema_stack(start_with: str) -> str:
    """Extract a formatted traceback for xmlschema package from current stack frame."""
    formatted_stack = traceback.format_stack()
    for k, line in enumerate(formatted_stack):
        if start_with in line:
            return ''.join(formatted_stack[k:])
    else:
        return ''.join(formatted_stack)


def dump_data(*args: Any) -> None:
    """Dump data to logger for debugging purposes."""
    if not args:
        return

    logging.basicConfig()
    chunks: list[str] = [' dump data for xmlschema debugging\n']

    for item in args:
        if isinstance(item, XMLResource):
            chunks.append(repr(item))
            if item.name:
                chunks.append(f'name: {item.name!r}')
            chunks.append(f'namespace: {item.namespace!r}')
            if item.url:
                chunks.append(f'URL: {item.url}')
            if not item.is_lazy():
                chunks.append(item.tostring())
            chunks.append('')
        else:
            chunks.append(repr(item))
            chunks.append('')

    logger.warning('\n'.join(chunks))



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import copy
from abc import abstractmethod
from collections import Counter
from collections.abc import Callable, ItemsView, Iterator, Mapping, ValuesView, Iterable
from operator import attrgetter
from types import MappingProxyType
from typing import Any, cast, NamedTuple, Optional, Union, TypeVar
from xml.etree.ElementTree import Element

import xmlschema.names as nm
from xmlschema.aliases import BaseXsdType, ElementType, LoadedItemType, \
    SchemaType, StagedItemType, SchemaGlobalType
from xmlschema.exceptions import XMLSchemaAttributeError, XMLSchemaKeyError, \
    XMLSchemaTypeError, XMLSchemaValueError
from xmlschema.translation import gettext as _
from xmlschema.utils.qnames import local_name, get_qname

from .helpers import parse_xsd_derivation
from .exceptions import XMLSchemaCircularityError, XMLSchemaModelDepthError
from .xsdbase import XsdComponent, XsdAnnotation
from .builtins import BUILTIN_TYPES
from .facets import XsdFacet, FACETS_CLASSES, XSD_10_FACETS, XSD_11_FACETS, \
    XSD_11_LIST_FACETS, XSD_10_LIST_FACETS, XSD_11_UNION_FACETS, XSD_10_UNION_FACETS
from .identities import XsdIdentity, XsdUnique, XsdKey, XsdKeyref, Xsd11Unique, \
    Xsd11Key, Xsd11Keyref
from .simple_types import XsdSimpleType, XsdAtomicBuiltin, XsdAtomicRestriction, \
    Xsd11AtomicRestriction, XsdUnion, Xsd11Union, XsdList

from .notations import XsdNotation
from .attributes import XsdAttribute, Xsd11Attribute, XsdAttributeGroup
from .complex_types import XsdComplexType, Xsd11ComplexType
from .wildcards import XsdAnyElement, Xsd11AnyElement, XsdAnyAttribute, Xsd11AnyAttribute
from .groups import XsdGroup, Xsd11Group
from .elements import XsdElement, Xsd11Element
from .assertions import XsdAssert

CT = TypeVar('CT', bound=XsdComponent)

BuilderType = Callable[[ElementType, SchemaType, Optional[XsdComponent]], CT]

# Elements for building dummy groups
ANY_ATTRIBUTE_ATTRIB = {'namespace': '##any', 'processContents': 'lax'}
ANY_ATTRIB = {
    'namespace': '##any',
    'processContents': 'lax',
    'minOccurs': '0',
    'maxOccurs': 'unbounded'
}

GLOBAL_MAP_INDEX = MappingProxyType({
    nm.XSD_SIMPLE_TYPE: 0,
    nm.XSD_COMPLEX_TYPE: 0,
    nm.XSD_NOTATION: 1,
    nm.XSD_ATTRIBUTE: 2,
    nm.XSD_ATTRIBUTE_GROUP: 3,
    nm.XSD_ELEMENT: 4,
    nm.XSD_GROUP: 5,
})

GLOBAL_MAP_ATTRIBUTE = MappingProxyType({
    nm.XSD_SIMPLE_TYPE: attrgetter('types'),
    nm.XSD_COMPLEX_TYPE: attrgetter('types'),
    nm.XSD_ATTRIBUTE: attrgetter('attributes'),
    nm.XSD_ATTRIBUTE_GROUP: attrgetter('attribute_groups'),
    nm.XSD_NOTATION: attrgetter('notations'),
    nm.XSD_ELEMENT: attrgetter('elements'),
    nm.XSD_GROUP: attrgetter('groups'),
})


class XsdBuilders:
    """
    A descriptor that is bound to a schema class for providing versioned builders
    for XSD components.
    """
    components: dict[str, type[XsdComponent]]
    facets: dict[str, type[XsdFacet]]
    identities: dict[str, type[XsdIdentity]]
    simple_types: dict[str, type[XsdSimpleType]]
    local_types: dict[str, Union[type[BaseXsdType], BuilderType[XsdSimpleType]]]
    builtins: tuple[dict[str, Any], ...]

    __slots__ = ('_name', '_xsd_version', 'components', 'facets', 'identities',
                 'simple_types', 'local_types', 'builtins', 'simple_type_class',
                 'notation_class', 'attribute_group_class', 'complex_type_class',
                 'attribute_class', 'group_class', 'element_class', 'any_element_class',
                 'any_attribute_class', 'atomic_restriction_class', 'list_class',
                 'union_class', 'unique_class', 'key_class', 'keyref_class',
                 'annotation_class', 'admitted_facets', 'admitted_union_facets',
                 'admitted_list_facets')

    def __init__(self, xsd_version: Optional[str] = None,
                 *facets_classes: type[XsdFacet],
                 **classes: type[XsdComponent]) -> None:
        if xsd_version is not None:
            self._xsd_version = xsd_version

        self.components = {}
        self.facets = {}

        if facets_classes:
            for cls in facets_classes:
                self.facets[cls.meta_tag()] = self.components[cls.meta_tag()] = cls

        for k, v in classes.items():
            if k.endswith('_class'):
                setattr(self, k, v)

    def __set_name__(self, cls: type[SchemaType], name: str) -> None:
        self._name = name
        self._xsd_version = getattr(cls, 'XSD_VERSION', '1.0')

        if not self.facets:
            self.facets.update(FACETS_CLASSES[self._xsd_version])
        else:
            facets = FACETS_CLASSES[self._xsd_version].copy()
            facets.update(self.facets)
            self.facets = facets

        self.builtins = BUILTIN_TYPES[self._xsd_version]

        self.simple_type_class = XsdSimpleType
        self.notation_class = XsdNotation
        self.attribute_group_class = XsdAttributeGroup
        self.list_class = XsdList
        self.annotation_class = XsdAnnotation

        if self._xsd_version == '1.0':
            self.complex_type_class = XsdComplexType
            self.attribute_class = XsdAttribute
            self.group_class = XsdGroup
            self.element_class = XsdElement
            self.any_element_class = XsdAnyElement
            self.any_attribute_class = XsdAnyAttribute
            self.atomic_restriction_class = XsdAtomicRestriction
            self.union_class = XsdUnion
            self.unique_class = XsdUnique
            self.key_class = XsdKey
            self.keyref_class = XsdKeyref
            self.admitted_facets = XSD_10_FACETS
            self.admitted_union_facets = XSD_10_UNION_FACETS
            self.admitted_list_facets = XSD_10_LIST_FACETS
        else:
            self.complex_type_class = Xsd11ComplexType
            self.attribute_class = Xsd11Attribute
            self.group_class = Xsd11Group
            self.element_class = Xsd11Element
            self.any_element_class = Xsd11AnyElement
            self.any_attribute_class = Xsd11AnyAttribute
            self.atomic_restriction_class = Xsd11AtomicRestriction
            self.union_class = Xsd11Union
            self.unique_class = Xsd11Unique
            self.key_class = Xsd11Key
            self.keyref_class = Xsd11Keyref
            self.admitted_facets = XSD_11_FACETS
            self.admitted_union_facets = XSD_11_UNION_FACETS
            self.admitted_list_facets = XSD_11_LIST_FACETS

        self.identities = {
            nm.XSD_UNIQUE: self.unique_class,
            nm.XSD_KEY: self.key_class,
            nm.XSD_KEYREF: self.keyref_class,
        }
        self.simple_types = {
            nm.XSD_RESTRICTION: self.atomic_restriction_class,
            nm.XSD_LIST: self.list_class,
            nm.XSD_UNION: self.union_class,
        }
        self.local_types = {
            nm.XSD_COMPLEX_TYPE: self.complex_type_class,
            nm.XSD_SIMPLE_TYPE: self.simple_type_factory,
        }

    def __setattr__(self, name: str, value: Any) -> None:
        if name == '_xsd_version':
            if value not in ('1.0', '1.1'):
                raise XMLSchemaValueError(f"wrong or unsupported XSD version {value!r}")
            elif hasattr(self, '_xsd_version') and self._xsd_version != value:
                raise XMLSchemaValueError("XSD version mismatch")

        elif name.endswith('_class'):
            if not isinstance(value, type) or not issubclass(value, XsdComponent):
                raise XMLSchemaTypeError(f"{name} must be a subclass of XsdComponent")
            if hasattr(self, name):
                return  # Skip changing a component class already set at __init__
            self.components[value.meta_tag()] = value

        super().__setattr__(name, value)

    def __get__(self, instance: Optional[Any], cls: type[Any]) -> 'XsdBuilders':
        return self

    def __set__(self, instance: Any, value: Any) -> None:
        raise XMLSchemaAttributeError(_("Can't set attribute {}").format(self._name))

    def __delete__(self, instance: Any) -> None:
        raise XMLSchemaAttributeError(_("Can't delete attribute {}").format(self._name))

    @property
    def xsd_version(self) -> str:
        return self._xsd_version

    def create_any_content_group(self, parent: Union[XsdComplexType, XsdGroup],
                                 any_element: Optional[XsdAnyElement] = None) -> XsdGroup:
        """
        Creates a local child model group for a complex type or a group that accepts any content.

        :param parent: the parent complex type or group for the content group.
        :param any_element: an optional any element to use for the content group. \
        When provided it's copied, linked to the group and the minOccurs/maxOccurs \
        are set to 0 and 'unbounded'.
        """
        schema = parent.schema
        elem = Element(nm.XSD_SEQUENCE)
        if isinstance(any_element, XsdAnyElement):
            attrib = any_element.elem.attrib.copy()
            attrib['minOccurs'] = '0'
            attrib['maxOccurs'] = 'unbounded'
            elem.append(Element(nm.XSD_ANY, attrib))
        else:
            elem.append(Element(nm.XSD_ANY, ANY_ATTRIB))

        elem.text = elem[0].tail = '\n  '
        return self.group_class(elem, schema, parent)

    def create_empty_content_group(self, parent: Union[XsdComplexType, XsdGroup],
                                   model: str = 'sequence', **attrib: Any) -> XsdGroup:
        """
        Creates an empty local child content group for a complex type or a group.
        """
        if model == 'sequence':
            elem = Element(nm.XSD_SEQUENCE, attrib)
        elif model == 'choice':
            elem = Element(nm.XSD_CHOICE, attrib)
        elif model == 'all':
            elem = Element(nm.XSD_ALL, attrib)
        else:
            msg = _("'model' argument must be (sequence | choice | all)")
            raise XMLSchemaValueError(msg)

        elem.text = '\n    '
        return self.group_class(elem, parent.schema, parent)

    def create_any_attribute_group(self, parent: Union[XsdComplexType, XsdElement]) \
            -> XsdAttributeGroup:
        """
        Creates a local child attribute group for a complex type or an element
        that accepts any attribute.
        """
        elem = Element(nm.XSD_ATTRIBUTE_GROUP)
        elem.append(Element(nm.XSD_ANY_ATTRIBUTE, ANY_ATTRIBUTE_ATTRIB))
        elem.text = elem[0].tail = '\n    '

        attribute_group = self.attribute_group_class(elem, parent.schema, parent)
        attribute_group[None] = self.any_attribute_class(
            elem[0], parent.schema, attribute_group
        )
        return attribute_group

    def create_empty_attribute_group(self, parent: Union[XsdComplexType, XsdElement]) \
            -> XsdAttributeGroup:
        """
        Creates an empty local child attribute group for a complex type or an element.
        """
        return self.attribute_group_class(
            Element(nm.XSD_ATTRIBUTE_GROUP), parent.schema, parent
        )

    def create_any_type(self, schema: SchemaType) -> XsdComplexType:
        """
        Creates a xs:anyType equivalent type related with the wildcards
        connected to global maps of the schema instance in order to do a
        correct namespace lookup during wildcards validation.
        """
        maps = schema.maps
        if schema.meta_schema is not None and schema.target_namespace != nm.XSD_NAMESPACE:
            schema = schema.meta_schema

        elem = Element(nm.XSD_COMPLEX_TYPE, name=nm.XSD_ANY_TYPE)
        elem.append(Element(nm.XSD_SEQUENCE))
        elem[0].append(Element(nm.XSD_ANY, ANY_ATTRIB))
        elem.append(Element(nm.XSD_ANY_ATTRIBUTE, ANY_ATTRIBUTE_ATTRIB))
        elem.text = elem[0].tail = '\n  '
        elem[0].text = elem[0][0].tail = '\n    '

        any_type = self.complex_type_class(elem, schema, mixed=True, block='', final='')
        for c in any_type.iter_components():
            c.maps = maps
        return any_type

    def create_element(self, name: str,
                       schema: SchemaType,
                       parent: Optional[XsdComponent] = None,
                       text: Optional[str] = None, **attrib: Any) -> XsdElement:
        """
        Creates a xs:element instance related to schema component.
        Used as dummy element for validation/decoding/encoding
        operations of wildcards and complex types.
        """
        elem = Element(nm.XSD_ELEMENT, name=name, **attrib)
        if text is not None:
            elem.text = text
        return self.element_class(elem, schema, parent)

    def simple_type_factory(self, elem: Element,
                            schema: SchemaType,
                            parent: Optional[XsdComponent] = None) -> XsdSimpleType:
        """
        Factory function for XSD simple types. Parses the xs:simpleType element and its
        child component, that can be a restriction, a list or a union. Annotations are
        linked to simple type instance, omitting the inner annotation if both are given.
        """
        annotation: Optional[XsdAnnotation] = None
        try:
            child = elem[0]
        except IndexError:
            return cast(XsdSimpleType, schema.maps.types[nm.XSD_ANY_SIMPLE_TYPE])
        else:
            if child.tag == nm.XSD_ANNOTATION:
                annotation = XsdAnnotation(child, schema, parent)
                try:
                    child = elem[1]
                except IndexError:
                    msg = _("(restriction | list | union) expected")
                    schema.parse_error(msg, elem)
                    return cast(XsdSimpleType, schema.maps.types[nm.XSD_ANY_SIMPLE_TYPE])

        xsd_type: XsdSimpleType
        try:
            xsd_type = self.simple_types[child.tag](child, schema, parent)
        except KeyError:
            msg = _("(restriction | list | union) expected")
            schema.parse_error(msg, elem)
            return cast(XsdSimpleType, schema.maps.types[nm.XSD_ANY_SIMPLE_TYPE])

        if annotation is not None:
            setattr(xsd_type, 'annotation', annotation)

        try:
            xsd_type.name = get_qname(schema.target_namespace, elem.attrib['name'])
        except KeyError:
            if parent is None:
                msg = _("missing attribute 'name' in a global simpleType")
                schema.parse_error(msg, elem)
                xsd_type.name = 'nameless_%s' % str(id(xsd_type))
        else:
            if parent is not None:
                msg = _("attribute 'name' not allowed for a local simpleType")
                schema.parse_error(msg, elem)
                xsd_type.name = None

        if 'final' in elem.attrib:
            xsd_type._final = parse_xsd_derivation(elem, 'final', validator=xsd_type)

        return xsd_type


class StagedMap(Mapping[str, CT]):
    label = 'component'

    @abstractmethod
    def _factory_or_class(self, elem: ElementType, schema: SchemaType) -> CT:
        """Returns the builder class or method used to build the global map."""

    __slots__ = ('_store', '_staging', '_builders')

    def __init__(self, builders: XsdBuilders):
        self._store: dict[str, CT] = {}
        self._staging: dict[str, StagedItemType] = {}
        self._builders = builders

    def __getitem__(self, qname: str) -> CT:
        try:
            return self._store[qname]
        except KeyError:
            if qname in self._staging:
                return self._build_global(qname)

            msg = _('global {} {!r} not found').format(self.label, qname)
            raise XMLSchemaKeyError(msg) from None

    def __iter__(self) -> Iterator[str]:
        yield from self._store

    def __len__(self) -> int:
        return len(self._store)

    def __repr__(self) -> str:
        return repr(self._store)

    def __copy__(self) -> 'StagedMap[CT]':
        obj = object.__new__(self.__class__)
        obj._builders = self._builders
        obj._staging = self._staging.copy()
        obj._store = self._store.copy()
        return obj

    copy = __copy__

    def clear(self) -> None:
        self._store.clear()
        self._staging.clear()

    def update(self, other: 'StagedMap[CT]') -> None:
        self._store.update(other._store)

    @property
    def total_staged(self) -> int:
        return len(self._staging)

    @property
    def staged(self) -> list[str]:
        return list(self._staging)

    @property
    def staged_items(self) -> ItemsView[str, StagedItemType]:
        return self._staging.items()

    @property
    def staged_values(self) -> ValuesView[StagedItemType]:
        return self._staging.values()

    def load(self, qname: str, elem: ElementType, schema: SchemaType) -> None:
        if qname in self._store:
            comp = self._store[qname]
            if comp.schema is schema:
                msg = _("global xs:{} with name={!r} is already built")
            elif comp.schema.maps is schema.maps or comp.schema.meta_schema is None:
                msg = _("global xs:{} with name={!r} is already defined")
            else:
                # Allows rebuilding of parent maps components for descendant maps
                # but not allows substitution of meta-schema components.
                self._staging[qname] = elem, schema
                return

        elif qname in self._staging:
            obj = self._staging[qname]

            if len(obj) == 2 and isinstance(obj, tuple):
                _elem, _schema = obj  # type: ignore[unused-ignore, misc]
                if _elem is elem and _schema is schema:
                    return  # ignored: it's the same component
                elif schema is _schema.override:
                    return  # ignored: the loaded component is overridden
                elif schema.override is _schema:
                    # replaced: the loaded component is an override
                    self._staging[qname] = (elem, schema)
                    return
                elif schema.meta_schema is None and _schema.meta_schema is not None:
                    return  # ignore merged meta-schema components
                elif _schema.meta_schema is None and schema.meta_schema is not None:
                    # Override merged meta-schema component
                    self._staging[qname] = (elem, schema)
                    return

            msg = _("global xs:{} with name={!r} is already loaded")
        else:
            self._staging[qname] = elem, schema
            return

        schema.parse_error(
            error=msg.format(local_name(elem.tag), qname),
            elem=elem
        )

    def load_redefine(self, qname: str, elem: ElementType, schema: SchemaType) -> None:
        try:
            item = self._staging[qname]
        except KeyError:
            schema.parse_error(_("not a redefinition!"), elem)
        else:
            if isinstance(item, list):
                item.append((elem, schema))
            else:
                self._staging[qname] = [cast(LoadedItemType, item), (elem, schema)]

    def load_override(self, qname: str, elem: ElementType, schema: SchemaType) -> None:
        if qname not in self._staging:
            # Overrides that match nothing in the target schema are ignored. See the
            # period starting with "Source declarations not present in the target set"
            # of the paragraph https://www.w3.org/TR/xmlschema11-1/#override-schema.
            return

        self._staging[qname] = elem, schema

    def build(self) -> None:
        for name in [x for x in self._staging]:
            if name in self._staging:
                self._build_global(name)

    def _build_global(self, qname: str) -> CT:
        obj = self._staging[qname]
        if isinstance(obj, tuple):
            # Not built XSD global component without redefinitions
            try:
                elem, schema = obj  # type: ignore[misc]
            except ValueError:
                raise XMLSchemaCircularityError(qname, *obj[0])

            # Encapsulate into a tuple to catch circular builds
            self._staging[qname] = ((elem, schema),)
            self._store[qname] = self._factory_or_class(elem, schema)
            self._staging.pop(qname)
            return self._store[qname]

        else:
            # Not built XSD global component with redefinitions
            try:
                elem, schema = obj[0]
            except ValueError:
                if not isinstance(obj, tuple):
                    raise
                raise XMLSchemaCircularityError(qname, *obj[0][0])

            self._staging[qname] = obj[0],  # To catch circular builds
            self._store[qname] = component = self._factory_or_class(elem, schema)
            self._staging.pop(qname)

            # Apply redefinitions (changing elem involve reparse of the component)
            for elem, schema in obj[1:]:
                if component.schema.target_namespace != schema.target_namespace:
                    msg = _("redefined schema {!r} has a different targetNamespace")
                    raise XMLSchemaValueError(msg.format(schema))

                component.redefine = copy.copy(component)
                component.redefine.parent = component
                component.schema = schema
                component.parse(elem)

            return self._store[qname]


class TypesMap(StagedMap[BaseXsdType]):

    def _factory_or_class(self, elem: ElementType, schema: SchemaType) -> BaseXsdType:
        if elem.tag == nm.XSD_COMPLEX_TYPE:
            return self._builders.complex_type_class(elem, schema)
        else:
            return self._builders.simple_type_factory(elem, schema)

    def build_builtins(self, schema: SchemaType) -> None:
        if schema.meta_schema is not None and nm.XSD_ANY_TYPE in self._store:
            # builtin types already provided
            return
        #
        # Special builtin types.
        #
        # xs:anyType
        # Ref: https://www.w3.org/TR/xmlschema11-1/#builtin-ctd
        self._store[nm.XSD_ANY_TYPE] = self._builders.create_any_type(schema)

        # xs:anySimpleType
        # Ref: https://www.w3.org/TR/xmlschema11-2/#builtin-stds
        any_simple_type = self._store[nm.XSD_ANY_SIMPLE_TYPE] = XsdSimpleType(
            elem=Element(nm.XSD_SIMPLE_TYPE, name=nm.XSD_ANY_SIMPLE_TYPE),
            schema=schema,
            parent=None,
            name=nm.XSD_ANY_SIMPLE_TYPE
        )

        # xs:anyAtomicType
        # Ref: https://www.w3.org/TR/xmlschema11-2/#builtin-stds
        self._store[nm.XSD_ANY_ATOMIC_TYPE] = \
            self._builders.atomic_restriction_class(
                elem=Element(nm.XSD_SIMPLE_TYPE, name=nm.XSD_ANY_ATOMIC_TYPE),
                schema=schema,
                parent=None,
                name=nm.XSD_ANY_ATOMIC_TYPE,
                base_type=any_simple_type,
            )

        for item in self._builders.builtins:
            item = item.copy()
            name: str = item['name']
            try:
                value = self._staging.pop(name)
            except KeyError:
                # If builtin type element is missing create a dummy element. Necessary for the
                # meta-schema XMLSchema.xsd of XSD 1.1, that not includes builtins declarations.
                elem = Element(nm.XSD_SIMPLE_TYPE, name=name, id=name)
            else:
                if not isinstance(value, tuple) or len(value) != 2:
                    continue
                elem, schema = value

            base_type: Optional[BaseXsdType]
            if 'base_type' in item:
                base_type = item['base_type'] = self._store[item['base_type']]
            else:
                base_type = None

            facets = item.pop('facets', None)
            xsd_type = XsdAtomicBuiltin(elem, schema, **item)
            if isinstance(facets, Iterable):
                built_facets = xsd_type.facets
                for e in facets:
                    try:
                        cls = self._builders.facets[e.tag]
                    except AttributeError:
                        built_facets[None] = e
                    else:
                        built_facets[e.tag] = cls(e, schema, xsd_type, base_type)
                xsd_type.facets = built_facets

            self._store[name] = xsd_type


class NotationsMap(StagedMap[XsdNotation]):
    label = 'notation'

    def _factory_or_class(self, elem: ElementType, schema: SchemaType) -> XsdNotation:
        return self._builders.notation_class(elem, schema)


class AttributesMap(StagedMap[XsdAttribute]):
    label = 'attribute'

    def _factory_or_class(self, elem: ElementType, schema: SchemaType) -> XsdAttribute:
        return self._builders.attribute_class(elem, schema)


class AttributeGroupsMap(StagedMap[XsdAttributeGroup]):
    label = 'attribute group'

    def _factory_or_class(self, elem: ElementType, schema: SchemaType) -> XsdAttributeGroup:
        return self._builders.attribute_group_class(elem, schema)


class ElementsMap(StagedMap[XsdElement]):
    label = 'element'

    def _factory_or_class(self, elem: ElementType, schema: SchemaType) -> XsdElement:
        return self._builders.element_class(elem, schema)


class GroupsMap(StagedMap[XsdGroup]):
    label = 'model group'

    def _factory_or_class(self, elem: ElementType, schema: SchemaType) -> XsdGroup:
        return self._builders.group_class(elem, schema)


class GlobalMaps(NamedTuple):
    types: TypesMap
    notations: NotationsMap
    attributes: AttributesMap
    attribute_groups: AttributeGroupsMap
    elements: ElementsMap
    groups: GroupsMap

    @classmethod
    def from_builders(cls, builders: XsdBuilders) -> 'GlobalMaps':
        return cls(
            TypesMap(builders),
            NotationsMap(builders),
            AttributesMap(builders),
            AttributeGroupsMap(builders),
            ElementsMap(builders),
            GroupsMap(builders)
        )

    def clear(self) -> None:
        for item in self:
            item.clear()

    def update(self, other: 'GlobalMaps') -> None:
        for m1, m2 in zip(self, other):
            m1.update(m2)  # type: ignore[attr-defined]

    def copy(self) -> 'GlobalMaps':
        return GlobalMaps(*[m.copy() for m in self])  # type: ignore[arg-type]

    def iter_globals(self) -> Iterator[SchemaGlobalType]:
        for item in self:
            yield from item.values()

    def iter_staged(self) -> Iterator[StagedItemType]:
        for item in self:
            yield from item.staged_values

    @property
    def total(self) -> int:
        """Total number of global components, fully or partially built."""
        return sum(len(m) for m in self)

    @property
    def total_built(self) -> int:
        """Total number of fully built global components."""
        return sum(1 for c in self.iter_globals() if c.built)

    @property
    def total_unbuilt(self) -> int:
        """Total number of not built or partially built global components."""
        return sum(1 for c in self.iter_globals() if not c.built)

    @property
    def total_staged(self) -> int:
        """Total number of staged global components."""
        return sum(m.total_staged for m in self)

    def load(self, schemas: Iterable[SchemaType]) -> None:
        """Loads global XSD components for the given schemas."""
        redefinitions = []

        for schema in schemas:
            if schema.target_namespace:
                ns_prefix = f'{{{schema.target_namespace}}}'
            else:
                ns_prefix = ''

            for elem in schema.root:
                if (tag := elem.tag) in (nm.XSD_REDEFINE, nm.XSD_OVERRIDE):
                    location = elem.get('schemaLocation')
                    if location is None:
                        continue

                    for child in elem:
                        try:
                            qname = ns_prefix + child.attrib['name']
                        except KeyError:
                            continue

                        try:
                            redefinitions.append(
                                (qname, elem, child, schema, schema.includes[location])
                            )
                        except KeyError:
                            if schema.partial:
                                redefinitions.append((qname, elem, child, schema, schema))

                elif tag in nm.GLOBAL_TAGS:
                    try:
                        qname = ns_prefix + elem.attrib['name']
                    except KeyError:
                        continue  # Invalid global: skip

                    self[GLOBAL_MAP_INDEX[tag]].load(qname, elem, schema)

        redefined_names = Counter(x[0] for x in redefinitions)
        for qname, elem, child, schema, redefined_schema in reversed(redefinitions):

            # Checks multiple redefinitions
            if redefined_names[qname] > 1:
                redefined_names[qname] = 1

                redefined_schemas: Any
                redefined_schemas = [x[-1] for x in redefinitions if x[0] == qname]
                if any(redefined_schemas.count(x) > 1 for x in redefined_schemas):
                    msg = _("multiple redefinition for {} {!r}")
                    schema.parse_error(
                        error=msg.format(local_name(child.tag), qname),
                        elem=child
                    )
                else:
                    redefined_schemas = {x[-1]: x[-2] for x in redefinitions if x[0] == qname}
                    for rs, s in redefined_schemas.items():
                        while True:
                            try:
                                s = redefined_schemas[s]
                            except KeyError:
                                break

                            if s is rs:
                                msg = _("circular redefinition for {} {!r}")
                                schema.parse_error(
                                    error=msg.format(local_name(child.tag), qname),
                                    elem=child
                                )
                                break

            if elem.tag == nm.XSD_REDEFINE:
                self[GLOBAL_MAP_INDEX[child.tag]].load_redefine(qname, child, schema)
            else:
                self[GLOBAL_MAP_INDEX[child.tag]].load_override(qname, child, schema)

    def build(self, schemas: Iterable[SchemaType]) -> None:
        """Builds global XSD components for the given schemas."""
        self.notations.build()
        self.attributes.build()
        self.attribute_groups.build()

        for schema in schemas:
            if not isinstance(schema.default_attributes, str):
                continue

            try:
                attributes = schema.maps.attribute_groups[schema.default_attributes]
            except KeyError:
                schema.default_attributes = None
                msg = _("defaultAttributes={0!r} doesn't match any attribute group of {1!r}")
                schema.parse_error(
                    error=msg.format(schema.root.get('defaultAttributes'), schema),
                    elem=schema.root
                )
            else:
                schema.default_attributes = attributes

        self.types.build()
        self.elements.build()
        self.groups.build()

        # Build element declarations inside model groups.
        for schema in schemas:
            for group in schema.iter_components(XsdGroup):
                try:
                    group.build()
                except XMLSchemaModelDepthError as e:
                    schema.parse_error(error=e, elem=group.elem)

        # Build identity references and XSD 1.1 assertions
        for schema in schemas:
            for obj in schema.iter_components((XsdIdentity, XsdAssert)):
                obj.build()



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
# mypy: ignore-errors
"""
This module contains abstact base class and helper
functions for building XSD based code generators.
"""
import os
import re
import sys
import inspect
import logging
from abc import ABC, ABCMeta
from fnmatch import fnmatch
from pathlib import Path
from typing import Optional

from jinja2 import Environment, ChoiceLoader, FileSystemLoader, \
    TemplateNotFound, TemplateAssertionError
from elementpath import datatypes

import xmlschema
from xmlschema.validators import XsdType, XsdElement, XsdAttribute
from xmlschema.names import XSD_NAMESPACE


NCNAME_PATTERN = re.compile(r'^[^\d\W][\w.\-]*$')
QNAME_PATTERN = re.compile(
    r'^(?:(?P<prefix>[^\d\W][\w\-.\xb7\u0387\u06DD\u06DE]*):)?'
    r'(?P<local>[^\d\W][\w\-.\xb7\u0387\u06DD\u06DE]*)$',
)


def is_shell_wildcard(pathname):
    return '*' in pathname or '?' in pathname or '[' in pathname


def xsd_qname(name):
    return f'{{{XSD_NAMESPACE}}}{name}'


def filter_method(func):
    """Marks a method for registration as template filter."""
    func.is_filter = True
    return func


def test_method(func):
    """Marks a method for registration as template test."""
    func.is_test = True
    return func


logger = logging.getLogger('xmlschema-codegen')


class GeneratorMeta(ABCMeta):
    """Metaclass for creating code generators. Checks formal_language """

    def __new__(mcs, name, bases, attrs):
        module = sys.modules.get(attrs['__module__'])
        module_path = getattr(module, '__file__', os.getcwd())

        formal_language = None
        searchpaths = []
        builtin_types = {}

        for base in bases:
            if getattr(base, 'formal_language', None):
                if formal_language is None:
                    formal_language = base.formal_language
                elif formal_language != base.formal_language:
                    raise ValueError("ambiguous formal_language from base classes")

            if getattr(base, 'searchpaths', None):
                searchpaths.extend(base.searchpaths)
            if getattr(base, 'builtin_types', None):
                builtin_types.update(base.builtin_types)

        if 'formal_language' not in attrs:
            attrs['formal_language'] = formal_language
        elif formal_language and formal_language != attrs['formal_language']:
            raise ValueError("formal_language cannot be changed")

        try:
            for path in attrs['searchpaths']:
                if Path(path).is_absolute():
                    dirpath = Path(path)
                else:
                    dirpath = Path(module_path).parent.joinpath(path)

                if not dirpath.is_dir():
                    raise ValueError(f"path {str(path)!r} is not a directory!")
                searchpaths.append(dirpath)

        except (KeyError, TypeError):
            pass
        else:
            attrs['searchpaths'] = searchpaths

        try:
            for k, v in attrs['builtin_types'].items():
                builtin_types[xsd_qname(k)] = v
        except (KeyError, AttributeError):
            pass
        finally:
            attrs['builtin_types'] = builtin_types

        return type.__new__(mcs, name, bases, attrs)


class AbstractGenerator(ABC, metaclass=GeneratorMeta):
    """
    Abstract base class for code generators based on Jinja2 template engine.

    :param schema: the source or the instance of the XSD schema.
    :param searchpath: additional search path for custom templates. \
    If provided the search path has priority over searchpaths defined \
    in generator class.
    :param types_map: a dictionary with custom mapping for XSD types.
    """
    formal_language: Optional[str] = None
    """The formal language associated to the code generator (eg. Python)."""

    searchpaths: Optional[list[str]] = None
    """
    Directory paths for searching templates, specified with a list or a tuple.
    Each path must be provided as relative from the directory of the module
    where the class is defined. Extends the searchpath defined in base classes.
    """

    builtin_types = {
        'anyType': '',
        'anySimpleType': '',
    }
    """
    Translation map for XSD builtin types. Updates the builtin_types
    defined in base classes.
    """

    def __init__(self, schema, searchpath=None, types_map=None):
        if isinstance(schema, xmlschema.XMLSchemaBase):
            self.schema = schema
        else:
            self.schema = xmlschema.XMLSchema11(schema)

        file_loaders = []
        if searchpath:
            file_loaders.append(FileSystemLoader(searchpath))
        if self.searchpaths is not None:
            file_loaders.extend(
                FileSystemLoader(str(path)) for path in reversed(self.searchpaths)
            )
        if not file_loaders:
            raise ValueError("no search paths defined!")
        loader = ChoiceLoader(file_loaders) if len(file_loaders) > 1 else file_loaders[0]

        self.types_map = self.builtin_types.copy()
        if types_map:
            if not self.schema.target_namespace:
                self.types_map.update(types_map)
            else:
                ns_part = '{%s}' % self.schema.target_namespace
                self.types_map.update((ns_part + k, v) for k, v in types_map.items())

        self.filters = {}
        self.tests = {}
        for name in filter(lambda x: callable(getattr(self, x)), dir(self)):
            method = getattr(self, name)
            if inspect.isfunction(method):
                # static methods
                if getattr(method, 'is_filter', False):
                    self.filters[name] = method
                elif getattr(method, 'is_test', False):
                    self.tests[name] = method
            elif inspect.isroutine(method) and hasattr(method, '__func__'):
                # class and instance methods
                if getattr(method.__func__, 'is_filter', False):
                    self.filters[name] = method
                elif getattr(method.__func__, 'is_test', False):
                    self.tests[name] = method

        type_mapping_filter = f'{self.formal_language}_type'.lower().replace(' ', '_')
        if type_mapping_filter not in self.filters:
            self.filters[type_mapping_filter] = self.map_type

        self._env = Environment(loader=loader)
        self._env.filters.update(self.filters)
        self._env.tests.update(self.tests)

    def __repr__(self):
        if self.schema.url:
            return f'{self.__class__.__name__}(schema={self.schema.name!r})'
        return f'{self.__class__.__name__}(namespace={self.schema.target_namespace!r})'

    def list_templates(self, extensions=None, filter_func=None):
        return self._env.list_templates(extensions, filter_func)

    def matching_templates(self, name):
        return self._env.list_templates(filter_func=lambda x: fnmatch(x, name))

    def get_template(self, name, parent=None, global_vars=None):
        return self._env.get_template(name, parent, global_vars)

    def select_template(self, names, parent=None, global_vars=None):
        return self._env.select_template(names, parent, global_vars)

    def render(self, names, parent=None, global_vars=None):
        if isinstance(names, str):
            names = [names]
        elif not all(isinstance(x, str) for x in names):
            raise TypeError("'names' argument must contain only strings!")

        results = []
        for name in names:
            try:
                template = self._env.get_template(name, parent, global_vars)
            except TemplateNotFound as err:
                logger.debug("name %r: %s", name, str(err))
            except TemplateAssertionError as err:
                logger.warning("template %r: %s", name, str(err))
            else:
                results.append(template.render(schema=self.schema))
        return results

    def render_to_files(self, names, parent=None, global_vars=None, output_dir='.', force=False):
        if isinstance(names, str):
            names = [names]
        elif not all(isinstance(x, str) for x in names):
            raise TypeError("'names' argument must contain only strings!")

        template_names = []
        for name in names:
            if is_shell_wildcard(name):
                template_names.extend(self.matching_templates(name))
            else:
                template_names.append(name)

        output_dir = Path(output_dir)
        rendered = []

        for name in template_names:
            try:
                template = self._env.get_template(name, parent, global_vars)
            except TemplateNotFound as err:
                logger.debug("name %r: %s", name, str(err))
            except TemplateAssertionError as err:
                logger.warning("template %r: %s", name, str(err))
            else:
                output_file = output_dir.joinpath(Path(name).name).with_suffix('')
                if not force and output_file.exists():
                    continue

                result = template.render(schema=self.schema)
                logger.info("write file %r", str(output_file))
                with open(output_file, 'w') as fp:
                    fp.write(result)
                rendered.append(str(output_file))

        return rendered

    def map_type(self, obj):
        """
        Maps an XSD type to a type declaration of the target language.
        This method is registered as filter with a name dependant from
        the language name (eg. c_type).

        :param obj: an XSD type or another type-related declaration as \
        an attribute or an element.
        :return: an empty string for non-XSD objects.
        """
        if isinstance(obj, XsdType):
            xsd_type = obj
        elif isinstance(obj, (XsdAttribute, XsdElement)):
            xsd_type = obj.type
        else:
            return ''

        try:
            return self.types_map[xsd_type.name]
        except KeyError:
            try:
                return self.types_map[xsd_type.base_type.name]
            except (KeyError, AttributeError):
                if xsd_type.is_complex():
                    return self.types_map[xsd_qname('anyType')]
                else:
                    return self.types_map[xsd_qname('anySimpleType')]

    @staticmethod
    @filter_method
    def name(obj, unnamed='none'):
        """
        Get the unqualified name of the provided object. Invalid
        chars for identifiers are replaced by an underscore.

        :param obj: an XSD object or a named object or a string.
        :param unnamed: value for unnamed objects. Defaults to 'none'.
        :return: str
        """
        try:
            name = obj.local_name
        except AttributeError:
            try:
                obj = obj.name
            except AttributeError:
                pass

            if not isinstance(obj, str):
                return unnamed

            try:
                if obj[0] == '{':
                    _, name = obj.split('}')
                elif ':' in obj:
                    prefix, name = obj.split(':')
                    if NCNAME_PATTERN.match(prefix) is None:
                        return ''
                else:
                    name = obj
            except (IndexError, ValueError):
                return ''
        else:
            if not isinstance(name, str):
                return ''

        if NCNAME_PATTERN.match(name) is None:
            return unnamed
        return name.replace('.', '_').replace('-', '_')

    @filter_method
    def qname(self, obj, unnamed='none', sep='__'):
        """
        Get the QName of the provided object. Invalid chars for
        identifiers are replaced by an underscore.

        :param obj: an XSD object or a named object or a string.
        :param unnamed: value for unnamed objects. Defaults to 'none'.
        :param sep: the replacement for colon. Defaults to double underscore.
        :return: str
        """
        try:
            qname = obj.prefixed_name
        except AttributeError:
            try:
                obj = obj.name
            except AttributeError:
                pass

            if not isinstance(obj, str):
                return unnamed

            try:
                if obj[0] == '{':
                    namespace, local_name = obj[1:].split('}')
                    for prefix, uri in self.schema.namespaces.items():
                        if uri == namespace:
                            qname = f'{prefix}:{local_name}'
                            break
                    else:
                        qname = local_name
                else:
                    qname = obj
            except IndexError:
                return ''
            except ValueError:
                return unnamed

        if not qname or QNAME_PATTERN.match(qname) is None:
            return unnamed
        return qname.replace('.', '_').replace('-', '_').replace(':', sep)

    @filter_method
    def namespace(self, obj):
        """Get the namespace URI of the provided object."""
        try:
            namespace = obj.target_namespace
        except AttributeError:
            if isinstance(obj, datatypes.QName):
                return obj.namespace
            elif not isinstance(obj, str):
                return ''

            try:
                if obj[0] == '{':
                    namespace, _ = obj[1:].split('}')
                    return namespace
                elif ':' in obj:
                    prefix, _ = obj.split(':')
                    return self.schema.namespaces.get(prefix, '')
                else:
                    return ''
            except (IndexError, ValueError):
                return ''
        else:
            return namespace if isinstance(namespace, str) else ''

    @staticmethod
    @filter_method
    def type_name(obj, suffix=None, unnamed='none'):
        """
        Get the unqualified name of the XSD type. Invalid
        chars for identifiers are replaced by an underscore.

        :param obj: an instance of (XsdType|XsdAttribute|XsdElement).
        :param suffix: force a suffix. For default removes '_type' or 'Type' suffixes.
        :param unnamed: value for unnamed XSD types. Defaults to 'none'.
        :return: str
        """
        if isinstance(obj, XsdType):
            name = obj.local_name or unnamed
        elif isinstance(obj, (XsdElement, XsdAttribute)):
            name = obj.type.local_name or unnamed
        else:
            name = unnamed

        if name.endswith('Type'):
            name = name[:-4]
        elif name.endswith('_type'):
            name = name[:-5]

        if suffix:
            name = f'{name}{suffix}'

        return name.replace('.', '_').replace('-', '_')

    @staticmethod
    @filter_method
    def type_qname(obj, suffix=None, unnamed='none', sep='__'):
        """
        Get the unqualified name of the XSD type. Invalid
        chars for identifiers are replaced by an underscore.

        :param obj: an instance of (XsdType|XsdAttribute|XsdElement).
        :param suffix: force a suffix. For default removes '_type' or 'Type' suffixes.
        :param unnamed: value for unnamed XSD types. Defaults to 'none'.
        :param sep: the replacement for colon. Defaults to double underscore.
        :return: str
        """
        if isinstance(obj, XsdType):
            qname = obj.prefixed_name or unnamed
        elif isinstance(obj, (XsdElement, XsdAttribute)):
            qname = obj.type.prefixed_name or unnamed
        else:
            qname = unnamed

        if qname.endswith('Type'):
            qname = qname[:-4]
        elif qname.endswith('_type'):
            qname = qname[:-5]

        if suffix:
            qname = f'{qname}{suffix}'

        return qname.replace('.', '_').replace('-', '_').replace(':', sep)

    @staticmethod
    @filter_method
    def sort_types(xsd_types, accept_circularity=False):
        """
        Returns a sorted sequence of XSD types usable for building type declarations.

        :param xsd_types: a sequence with XSD types.
        :param accept_circularity: if set to `True` circularities \
        are accepted. Defaults to `False`.
        :return: a list with ordered types.
        """
        if not isinstance(xsd_types, (list, tuple)):
            try:
                xsd_types = list(xsd_types.values())
            except AttributeError:
                pass

        assert all(isinstance(x, XsdType) for x in xsd_types)
        ordered_types = [x for x in xsd_types if x.is_simple()]
        ordered_types.extend(x for x in xsd_types if x.is_complex() and x.has_simple_content())
        unordered = {x: [] for x in xsd_types if x.is_complex() and not x.has_simple_content()}

        for xsd_type in unordered:
            for e in xsd_type.content.iter_elements():
                if e.type in unordered:
                    unordered[xsd_type].append(e.type)

        while unordered:
            deleted = 0
            for xsd_type in xsd_types:
                if xsd_type in unordered:
                    if not unordered[xsd_type]:
                        del unordered[xsd_type]
                        ordered_types.append(xsd_type)
                        deleted += 1

            for xsd_type in unordered:
                unordered[xsd_type] = [x for x in unordered[xsd_type] if x in unordered]

            if not deleted:
                if not accept_circularity:
                    raise ValueError(f"circularity found between {list(unordered)!r}")
                ordered_types.extend(list(unordered))
                break

        assert len(xsd_types) == len(ordered_types)
        return ordered_types

    def is_derived(self, xsd_type, *names, derivation=None):
        """
        Returns `True` if the argument XSD type is derived from any
        of other types expressed by name, otherwise returns `False`.

        :param xsd_type: an XsdComplexType/XsdSimpleType instance.
        :param names: positional argument with the names of other \
        XSD types.
        :param derivation: the type of derivation, that can be \
        *extension* or *restriction*, or both with a space separator. \
        If no value is provided it only checks if it is derived from \
        or if it is the XSD type itself.
        """
        for type_name in names:
            if not isinstance(type_name, str) or not type_name:
                continue  # pragma: no cover
            elif type_name[0] == '{':
                other = self.schema.maps.types.get(type_name)
            else:
                try:
                    expanded_name = self.schema.resolve_qname(type_name)
                except xmlschema.XMLSchemaException:
                    other = self.schema.types.get(type_name)
                else:
                    other = self.schema.maps.types.get(expanded_name)
                    if other is None:
                        other = self.schema.types.get(type_name)

            if other is not None and xsd_type.is_derived(other, derivation):
                return True

        return False

    @test_method
    def derivation(self, xsd_type, *names):
        return self.is_derived(xsd_type, *names)

    @test_method
    def extension(self, xsd_type, *names):
        return self.is_derived(xsd_type, *names, derivation='extension')

    @test_method
    def restriction(self, xsd_type, *names):
        return self.is_derived(xsd_type, *names, derivation='restriction')

    @staticmethod
    @test_method
    def multi_sequence(xsd_type):
        try:
            return any(e.is_multiple() for e in xsd_type.content.iter_elements())
        except AttributeError:
            return False


class PythonGenerator(AbstractGenerator):
    """A Python code generator for XSD schemas."""

    formal_language = 'Python'

    searchpaths = ['templates/python/']

    builtin_types = {
        'string': 'str',
        'decimal': 'decimal.Decimal',
        'float': 'float',
        'double': 'float',

        'date': 'datatypes.Date10',
        'dateTime': 'datatypes.DateTime10',
        'gDay': 'datatypes.GregorianDay',
        'gMonth': 'datatypes.GregorianMonth',
        'gMonthDay': 'datatypes.GregorianMonthDay',
        'gYear': 'datatypes.GregorianYear10',
        'gYearMonth': 'datatypes.GregorianYearMonth10',
        'time': 'datatypes.Time',
        'duration': 'datatypes.Duration',

        'QName': 'datatypes.QName',
        'NOTATION': 'datatypes.DateTime10',
        'anyURI': 'datatypes.AnyURI',
        'boolean': 'bool',
        'base64Binary': 'datatypes.Base64Binary',
        'hexBinary': 'datatypes.HexBinary',
        'normalizedString': 'str',
        'token': 'str',
        'language': 'str',
        'Name': 'str',
        'NCName': 'str',
        'ID': 'str',
        'IDREF': 'str',
        'ENTITY': 'str',
        'NMTOKEN': 'str',

        'integer': 'int',
        'long': 'int',
        'int': 'int',
        'short': 'int',
        'byte': 'int',
        'nonNegativeInteger': 'int',
        'positiveInteger': 'int',
        'unsignedLong': 'int',
        'unsignedInt': 'int',
        'unsignedShort': 'int',
        'unsignedByte': 'int',
        'nonPositiveInteger': 'int',
        'negativeInteger': 'int',

        # XSD 1.1 built-in types
        'dateTimeStamp': 'datatypes.DateTimeStamp10',
        'dayTimeDuration': 'datatypes.DayTimeDuration',
        'yearMonthDuration': 'datatypes.YearMonthDuration',
    }



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
# mypy: ignore-errors
"""
Test factory for creating test cases from lists of paths to XSD or XML files.

The list of cases can be defined within files named "testfiles". These are text files
that contain a list of relative paths to XSD or XML files, that are used to dinamically
build a set of test classes. Each path is followed by a list of options that defines a
custom setting for each test.
"""
import re
import argparse
import os
import fileinput
import logging
import platform
import sys
import unittest

from xmlschema.cli import xsd_version_number, defuse_data
from xmlschema.validators import XMLSchema10, XMLSchema11
from ._observers import ObservedXMLSchema10, ObservedXMLSchema11

logger = logging.getLogger(__file__)


def get_test_args(args_line):
    """Returns the list of arguments from provided text line."""
    try:
        args_line, _ = args_line.split('#', 1)  # Strip optional ending comment
    except ValueError:
        pass
    return re.split(r'(?<!\\) ', args_line.strip())


def get_test_program_args_parser(prog=None):
    """
    Gets an argument parser for building test scripts for schemas and xml files.
    The returned parser has many arguments of unittest's TestProgram plus some
    arguments for selecting testfiles and XML schema options.
    """
    parser = argparse.ArgumentParser(os.path.basename(prog), add_help=True)
    default_testfiles = os.path.join(os.path.dirname(prog), 'test_cases/testfiles')

    # unittest's arguments
    parser.add_argument('-v', '--verbose', dest='verbosity', default=1,
                        action='store_const', const=2, help='Verbose output')
    parser.add_argument('-q', '--quiet', dest='verbosity',
                        action='store_const', const=0, help='Quiet output')
    parser.add_argument('--locals', dest='tb_locals', action='store_true',
                        help='Show local variables in tracebacks')
    parser.add_argument('-f', '--failfast', dest='failfast',
                        action='store_true', help='Stop on first fail or error')
    parser.add_argument('-c', '--catch', dest='catchbreak',
                        action='store_true', help='Catch Ctrl-C and display results so far')
    parser.add_argument('-b', '--buffer', dest='buffer', action='store_true',
                        help='Buffer stdout and stderr during tests')
    parser.add_argument('-k', dest='patterns', action='append', default=list(),
                        help='Only run tests which match the given substring')

    # xmlschema's arguments (removed by the test script)
    parser.add_argument('--lxml', dest='lxml', action='store_true', default=False,
                        help='Check also with lxml.etree.XMLSchema (for XSD 1.0)')
    parser.add_argument('--codegen', action="store_true", default=False,
                        help="Test code generation with XML data bindings module.")
    parser.add_argument('--random', dest='random', action='store_true', default=False,
                        help='Execute the test cases in random order.')
    parser.add_argument('testfiles', type=str, nargs='*', default=default_testfiles,
                        help="Test files containing a list of cases to build and run.")
    return parser


def parse_xmlschema_args(argv=None):
    """Parse CLI arguments removing xmlschema's additional arguments after parsing."""
    if argv is None:
        argv = sys.argv

    prog, args = argv[0], argv[1:]
    parser = get_test_program_args_parser(prog)
    args = parser.parse_args(args)

    # Clean argv of xmlschema arguments and
    _argv = sys.argv.copy()
    argv.clear()
    argv.append(prog)
    for item in _argv[1:]:
        if item.endswith('testfiles'):
            continue
        elif item not in ('--lxml', '--codegen', '--random', '-f',
                          '--failfast', '-c', '--catch', '-b', '--buffer'):
            argv.append(item)

    return args


def run_xmlschema_tests(target=None, args=None):
    if target is not None:
        header_template = "Test xmlschema {} with Python {} on platform {}"
        header = header_template.format(
            target, platform.python_version(), platform.platform()
        )
        print('{0}\n{1}\n{0}'.format("*" * len(header), header))

    if args is None:
        unittest.main()
    else:
        unittest.main(
            argv=sys.argv,
            verbosity=args.verbosity,
            failfast=args.failfast,
            catchbreak=args.catchbreak,
            buffer=args.buffer
        )


def get_test_line_args_parser():
    """Gets an arguments parser for uncommented on not blank "testfiles" lines."""

    parser = argparse.ArgumentParser(add_help=True)
    parser.usage = "TEST_FILE [OPTIONS]\nTry 'TEST_FILE --help' for more information."
    parser.add_argument('filename', metavar='TEST_FILE', type=str,
                        help="Test filename (relative path).")
    parser.add_argument(
        '-L', dest='locations', nargs=2, type=str, default=None, action='append',
        metavar="URI-URL", help="Schema location hint overrides."
    )
    parser.add_argument(
        '--version', dest='version', metavar='VERSION', type=xsd_version_number, default='1.0',
        help="XSD schema version to use for the test case (default is 1.0)."
    )
    parser.add_argument(
        '--errors', type=int, default=0, metavar='NUM',
        help="Number of errors expected (default=0)."
    )
    parser.add_argument(
        '--warnings', type=int, default=0, metavar='NUM',
        help="Number of warnings expected (default=0)."
    )
    parser.add_argument(
        '--inspect', action="store_true", default=False,
        help="Inspect using an observed custom schema class."
    )
    parser.add_argument(
        '--defuse', metavar='(always, remote, never)', type=defuse_data, default='remote',
        help="Define when to use the defused XML data loaders."
    )
    parser.add_argument(
        '--timeout', type=int, default=300, metavar='SEC',
        help="Timeout for fetching resources (default=300)."
    )
    parser.add_argument(
        '--validation-only', action="store_true", default=False,
        help="Skip decode/encode tests on XML data."
    )
    parser.add_argument(
        '--no-pickle', action="store_true", default=False,
        help="Skip pickling/unpickling test on schema (max recursion exceeded)."
    )
    parser.add_argument('--skip-location-loader', action="store_true", default=False,
                        help="Skip test with alternative LocationSchemaLoader.")
    parser.add_argument(
        '--lax-encode', action="store_true", default=False,
        help="Use lax mode on encode checks (for cases where test data uses default or "
             "fixed values or some test data are skipped by wildcards processContents). "
             "Ignored on schema tests."
    )
    parser.add_argument(
        '--debug', action="store_true", default=False,
        help="Activate the debug mode (only the cases with --debug are executed).",
    )
    parser.add_argument(
        '--codegen', action="store_true", default=False,
        help="Test code generation with XML data bindings module. For default "
             "test code generation if the same command option is provided.",
    )
    return parser


def xmlschema_tests_factory(test_class_builder, testfiles, suffix,
                            check_with_lxml=False, codegen=False, verbosity=1):
    """
    Factory function for file based schema/validation cases.

    :param test_class_builder: the test class builder function.
    :param testfiles: a single or a list of testfiles indexes.
    :param suffix: the suffix ('xml' or 'xsd') to consider for cases.
    :param check_with_lxml: if `True` compare with lxml XMLSchema class, \
    reporting anomalies. Works only for XSD 1.0 tests.
    :param codegen: if `True` is provided checks code generation with XML data \
    bindings module for all tests. For default is `False` and code generation \
    is tested only for the cases where the same option is provided.
    :param verbosity: the unittest's verbosity, can be 0, 1 or 2.
    :return: a list of test classes.
    """
    test_classes = {}
    test_num = 0
    debug_mode = False
    line_buffer = []
    test_line_parser = get_test_line_args_parser()

    for line in fileinput.input(testfiles):
        line = line.strip()
        if not line or line[0] == '#':
            if not line_buffer:
                continue
            else:
                raise SyntaxError("Empty continuation at line %d!" % fileinput.filelineno())
        elif '#' in line:
            line = line.split('#', 1)[0].rstrip()

        # Process line continuations
        if line[-1] == '\\':
            line_buffer.append(line[:-1].strip())
            continue
        elif line_buffer:
            line_buffer.append(line)
            line = ' '.join(line_buffer)
            del line_buffer[:]

        test_args = test_line_parser.parse_args(get_test_args(line))
        if test_args.locations is not None:
            test_args.locations = {k.strip('\'"'): v for k, v in test_args.locations}
        if codegen:
            test_args.codegen = True

        test_file = os.path.join(os.path.dirname(fileinput.filename()), test_args.filename)
        if os.path.isdir(test_file):
            logger.debug("Skip %s: is a directory.", test_file)
            continue
        elif os.path.splitext(test_file)[1].lower() != '.%s' % suffix:
            logger.debug("Skip %s: wrong suffix.", test_file)
            continue
        elif not os.path.isfile(test_file):
            logger.error("Skip %s: is not a file.", test_file)
            continue

        test_num += 1

        # Debug mode activation
        if debug_mode:
            if not test_args.debug:
                continue
        elif test_args.debug:
            debug_mode = True
            msg = "Debug mode activated: discard previous %r test classes."
            logger.debug(msg, len(test_classes))
            test_classes.clear()

        if test_args.version == '1.0':
            schema_class = ObservedXMLSchema10 if test_args.inspect else XMLSchema10
            test_class = test_class_builder(
                test_file, test_args, test_num, schema_class, check_with_lxml
            )
        else:
            schema_class = ObservedXMLSchema11 if test_args.inspect else XMLSchema11
            test_class = test_class_builder(
                test_file, test_args, test_num, schema_class, check_with_lxml=False
            )

        test_classes[test_class.__name__] = test_class
        if verbosity == 2:
            print(f"Create case {test_class.__name__} for file {os.path.relpath(test_file)}")

        logger.debug("Add XSD %s test class %r.", test_args.version, test_class.__name__)

    if line_buffer:
        raise ValueError("Not completed line continuation at the end!")

    return test_classes



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from decimal import Decimal
from math import isinf, isnan
from typing import Optional, SupportsInt, SupportsFloat, TYPE_CHECKING, Union
from xml.etree.ElementTree import Element
from elementpath import datatypes

import xmlschema.names as nm
from xmlschema.aliases import ElementType, SchemaType
from xmlschema.exceptions import XMLSchemaValueError
from xmlschema.translation import gettext as _
from .exceptions import XMLSchemaValidationError

if TYPE_CHECKING:
    from xmlschema.validators import XsdAnnotation, XsdComponent  # noqa: F401

XSD_FINAL_ATTRIBUTE_VALUES = {'restriction', 'extension', 'list', 'union'}
XSD_BOOLEAN_MAP = {
    'false': False, '0': False,
    'true': True, '1': True
}


def get_xsd_annotation(elem: ElementType,
                       schema: SchemaType,
                       parent: Optional['XsdComponent'] = None) -> Optional['XsdAnnotation']:
    """
    Returns the XSD annotation from the 1st child element of the provided element,
    `None` if it doesn't exist.
    """
    for child in elem:
        if child.tag == nm.XSD_ANNOTATION:
            return schema.builders.annotation_class(
                child, schema, parent, parent_elem=elem
            )
        elif not callable(child.tag):
            return None
    else:
        return None


def get_schema_annotations(schema: SchemaType) -> list['XsdAnnotation']:
    annotations = []
    annotation_class = schema.builders.annotation_class

    for elem in schema.source.root:
        if elem.tag == nm.XSD_ANNOTATION:
            annotations.append(annotation_class(elem, schema))
        elif elem.tag in nm.SCHEMA_DECLARATION_TAGS or elem.tag == nm.XSD_DEFAULT_OPEN_CONTENT:
            annotation = get_xsd_annotation(elem, schema)
            if annotation is not None:
                annotations.append(annotation)

    return annotations


def parse_xsd_derivation(elem: Element,
                         name: str,
                         choices: Union[None, set[str], tuple[str, ...]] = None,
                         validator: Union[None, SchemaType, 'XsdComponent'] = None) -> str:
    """
    Get a derivation attribute (maybe 'block', 'blockDefault', 'final' or 'finalDefault')
    checking the items with the values arguments. Returns a string.

    :param elem: the Element instance.
    :param name: the attribute name.
    :param choices: a set of admitted values when the attribute value is not '#all'.
    :param validator: optional schema or a component (element or type) to report \
    a parse error instead of raising a `ValueError`.
    """
    try:
        value = elem.attrib[name]
    except KeyError:
        return ''

    if choices is None:
        choices = XSD_FINAL_ATTRIBUTE_VALUES

    items = value.split()
    if len(items) == 1 and items[0] == '#all':
        return ' '.join(choices)
    elif not all(s in choices for s in items):
        msg = _("wrong value {!r} for attribute {!r}").format(value, name)
        if validator is None:
            raise ValueError(msg)
        validator.parse_error(msg)
        return ''
    return value


def parse_xpath_default_namespace(validator: Union[SchemaType, 'XsdComponent']) -> str:
    """
    Parse XSD 1.1 xpathDefaultNamespace attribute for schema, alternative, assert, assertion
    and selector declarations, checking if the value is conforming to the specification. In
    case the attribute is missing or for wrong attribute values defaults to ''.
    """
    try:
        value = validator.elem.attrib['xpathDefaultNamespace']
    except KeyError:
        return ''

    value = value.strip()
    if value == '##local':
        return ''
    elif value == '##defaultNamespace':
        return validator.default_namespace
    elif value == '##targetNamespace':
        return validator.target_namespace
    elif len(value.split()) == 1:
        return value
    else:
        admitted_values = ('##defaultNamespace', '##targetNamespace', '##local')
        msg = _("wrong value {0!r} for 'xpathDefaultNamespace' "
                "attribute, can be (anyURI | {1}).")
        validator.parse_error(msg.format(value, ' | '.join(admitted_values)))
        return ''


def parse_target_namespace(validator: Union[SchemaType, 'XsdComponent']) -> str:
    """
    XSD 1.1 targetNamespace attribute in schema, elements and attributes declarations.
    """
    try:
        target_namespace = validator.elem.attrib['targetNamespace'].strip()
    except KeyError:
        return ''

    if target_namespace == nm.XMLNS_NAMESPACE:
        # https://www.w3.org/TR/xmlschema11-1/#sec-nss-special
        msg = _(f"The namespace {nm.XMLNS_NAMESPACE} cannot be used as 'targetNamespace'")
        raise XMLSchemaValueError(msg)
    elif not target_namespace and validator.elem.tag == nm.XSD_SCHEMA:
        # https://www.w3.org/TR/2004/REC-xmlschema-1-20041028/structures.html#element-schema
        msg = _("the attribute 'targetNamespace' cannot be an empty string")
        validator.parse_error(msg)

    return target_namespace


#
# XSD built-in types validator functions

def decimal_validator(value: Union[Decimal, int, float, str]) -> None:
    try:
        if not isinstance(value, (Decimal, float)):
            datatypes.DecimalProxy.validate(value)
        elif isinf(value) or isnan(value):
            raise ValueError()
    except (ValueError, TypeError):
        raise XMLSchemaValidationError(decimal_validator, value,
                                       _("value is not a valid xs:decimal")) from None


def qname_validator(value: str) -> None:
    if datatypes.QName.pattern.match(value) is None:
        raise XMLSchemaValidationError(qname_validator, value,
                                       _("value is not an xs:QName"))


def byte_validator(value: int) -> None:
    if not (-2**7 <= value < 2 ** 7):
        raise XMLSchemaValidationError(int_validator, value,
                                       _("value must be {:s}").format("-128 <= x < 128"))


def short_validator(value: int) -> None:
    if not (-2**15 <= value < 2 ** 15):
        raise XMLSchemaValidationError(short_validator, value,
                                       _("value must be {:s}").format("-2^15 <= x < 2^15"))


def int_validator(value: int) -> None:
    if not (-2**31 <= value < 2 ** 31):
        raise XMLSchemaValidationError(int_validator, value,
                                       _("value must be {:s}").format("-2^31 <= x < 2^31"))


def long_validator(value: int) -> None:
    if not (-2**63 <= value < 2 ** 63):
        raise XMLSchemaValidationError(long_validator, value,
                                       _("value must be {:s}").format("-2^63 <= x < 2^63"))


def unsigned_byte_validator(value: int) -> None:
    if not (0 <= value < 2 ** 8):
        raise XMLSchemaValidationError(unsigned_byte_validator, value,
                                       _("value must be {:s}").format("0 <= x < 256"))


def unsigned_short_validator(value: int) -> None:
    if not (0 <= value < 2 ** 16):
        raise XMLSchemaValidationError(unsigned_short_validator, value,
                                       _("value must be {:s}").format("0 <= x < 2^16"))


def unsigned_int_validator(value: int) -> None:
    if not (0 <= value < 2 ** 32):
        raise XMLSchemaValidationError(unsigned_int_validator, value,
                                       _("value must be {:s}").format("0 <= x < 2^32"))


def unsigned_long_validator(value: int) -> None:
    if not (0 <= value < 2 ** 64):
        raise XMLSchemaValidationError(unsigned_long_validator, value,
                                       _("value must be {:s}").format("0 <= x < 2^64"))


def negative_int_validator(value: int) -> None:
    if value >= 0:
        raise XMLSchemaValidationError(negative_int_validator, value,
                                       _("value must be negative"))


def positive_int_validator(value: int) -> None:
    if value <= 0:
        raise XMLSchemaValidationError(positive_int_validator, value,
                                       _("value must be positive"))


def non_positive_int_validator(value: int) -> None:
    if value > 0:
        raise XMLSchemaValidationError(non_positive_int_validator, value,
                                       _("value must be non positive"))


def non_negative_int_validator(value: int) -> None:
    if value < 0:
        raise XMLSchemaValidationError(non_negative_int_validator, value,
                                       _("value must be non negative"))


def hex_binary_validator(value: Union[str, datatypes.HexBinary]) -> None:
    if not isinstance(value, datatypes.HexBinary) and \
            datatypes.HexBinary.pattern.match(value) is None:
        raise XMLSchemaValidationError(hex_binary_validator, value,
                                       _("not an hexadecimal number"))


def base64_binary_validator(value: Union[str, datatypes.Base64Binary]) -> None:
    if isinstance(value, datatypes.Base64Binary):
        return
    value = value.replace(' ', '')
    if not value:
        return

    match = datatypes.Base64Binary.pattern.match(value)
    if match is None or match.group(0) != value:
        raise XMLSchemaValidationError(base64_binary_validator, value,
                                       _("not a base64 encoding"))


def error_type_validator(value: object) -> None:
    raise XMLSchemaValidationError(error_type_validator, value,
                                   _("no value is allowed for xs:error type"))


#
# XSD builtin decoding functions

def boolean_to_python(value: str) -> bool:
    try:
        return XSD_BOOLEAN_MAP[value]
    except KeyError:
        raise XMLSchemaValueError(_('{!r} is not a boolean value').format(value))


def python_to_boolean(value: object) -> str:
    if isinstance(value, str):
        if value in XSD_BOOLEAN_MAP:
            return value
        raise XMLSchemaValueError(_('{!r} is not a boolean value').format(value))
    return str(value).lower()


def python_to_float(value: Union[SupportsFloat, str]) -> str:
    if isinstance(value, str):
        if value in ('NaN', 'INF', '-INF'):
            return value
        return str(float(value))
    elif isnan(value):
        return "NaN"
    if value == float("inf"):
        return "INF"
    if value == float("-inf"):
        return "-INF"
    return str(value)


def python_to_int(value: Union[SupportsInt, str]) -> str:
    return str(int(value))
